@@ -179,6 +179,8 @@ bool catalog_add_s_table_chksum(DatabaseCatalog *catalog,
 								TableChecksum *srcChk,
 								TableChecksum *dstChk);
 
+bool catalog_add_s_table_size(DatabaseCatalog *catalog,
+							  SourceTableSize *tableSize);
 bool catalog_delete_s_table(DatabaseCatalog *catalog,
 							const char *nspname,
 							const char *relname);
@@ -204,9 +206,6 @@ typedef struct SourceTableIterator
 	DatabaseCatalog *catalog;
 	SourceTable *table;
 	SQLiteQuery query;
-
-	/* optional parameters */
-	uint64_t splitTableLargerThanBytes;
 } SourceTableIterator;
 
 bool catalog_iter_s_table_init(SourceTableIterator *iter);
@@ -248,8 +247,11 @@ bool catalog_iter_s_table_part_init(SourceTablePartsIterator *iter);
 bool catalog_iter_s_table_part_next(SourceTablePartsIterator *iter);
 bool catalog_iter_s_table_part_finish(SourceTablePartsIterator *iter);
 
+bool catalog_s_table_attrlist(DatabaseCatalog *catalog, SourceTable *table);
 bool catalog_s_table_part_fetch(SQLiteQuery *query);
 
+bool catalog_s_table_fetch_attrlist(SQLiteQuery *query);
+
 typedef struct SourceTableAttrsIterator
 {
 	DatabaseCatalog *catalog;
@@ -342,6 +344,7 @@ bool catalog_s_index_fetch(SQLiteQuery *query);
  */
 bool catalog_add_s_seq(DatabaseCatalog *catalog, SourceSequence *index);
 bool catalog_update_sequence_values(DatabaseCatalog *catalog, SourceSequence *seq);
+bool catalog_update_s_table_relpages(DatabaseCatalog *catalog, SourceTable *sourceTable);
 
 typedef bool (SourceSequenceIterFun)(void *context, SourceSequence *seq);
 
@@ -372,7 +375,9 @@ bool catalog_s_seq_fetch(SQLiteQuery *query);
  * pg_restore archives TOC list names (restore_list_name) in such a way that we
  * can get away with a single hash-table like lookup.
  */
-bool catalog_prepare_filter(DatabaseCatalog *catalog);
+bool catalog_prepare_filter(DatabaseCatalog *catalog,
+							bool skipExtensions,
+							bool skipCollations);
 
 typedef struct CatalogFilter
 {
@@ -444,10 +449,12 @@ bool catalog_s_database_guc_fetch(SQLiteQuery *query);
  * Namespaces
  */
 bool catalog_add_s_namespace(DatabaseCatalog * catalog, SourceSchema *namespace);
-
-bool catalog_lookup_s_namespace_by_rlname(DatabaseCatalog *catalog,
-										  const char *restoreListName,
-										  SourceSchema *result);
+bool catalog_lookup_s_namespace_by_oid(DatabaseCatalog *catalog,
+									   uint32_t oid,
+									   SourceSchema *result);
+bool catalog_lookup_s_namespace_by_nspname(DatabaseCatalog *catalog,
+										   const char *nspname,
+										   SourceSchema *result);
 
 bool catalog_s_namespace_fetch(SQLiteQuery *query);
 
@@ -616,6 +623,8 @@ typedef struct BindParam
 } BindParam;
 
 
+bool catalog_execute(DatabaseCatalog *catalog, char *sql);
+
 bool catalog_sql_prepare(sqlite3 *db, const char *sql, SQLiteQuery *query);
 bool catalog_sql_bind(SQLiteQuery *query, BindParam *params, int count);
 bool catalog_sql_execute(SQLiteQuery *query);
