@@ -40,11 +40,13 @@
 	"  --no-owner                    Do not set ownership of objects to match the original database\n" \
 	"  --no-acl                      Prevent restoration of access privileges (grant/revoke commands).\n" \
 	"  --no-comments                 Do not output commands to restore comments\n" \
+	"  --no-tablespaces              Do not output commands to select tablespaces\n" \
 	"  --skip-large-objects          Skip copying large objects (blobs)\n" \
 	"  --skip-extensions             Skip restoring extensions\n" \
 	"  --skip-ext-comments           Skip restoring COMMENT ON EXTENSION\n" \
 	"  --skip-collations             Skip restoring collations\n" \
 	"  --skip-vacuum                 Skip running VACUUM ANALYZE\n" \
+	"  --skip-db-properties          Skip copying ALTER DATABASE SET properties\n" \
 	"  --requirements <filename>     List extensions requirements\n" \
 	"  --filters <filename>          Use the filters defined in <filename>\n" \
 	"  --fail-fast                   Abort early in case of error\n" \
@@ -248,6 +250,24 @@ clone_and_follow(CopyDataSpec *copySpecs)
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
+	/*
+	 * We fetch the schema here, rather than later in the clone subprocess,
+	 * which simply reuses this cached data. This is done to avoid lock
+	 * contention between the clone and follow subprocesses, as they both try to
+	 * write concurrently to the source.db SQLite database, leading one to
+	 * failure. This is also necessary for plugins like test_decoding, which
+	 * require information such as primary keys.
+	 *
+	 * In the future, if the follow subprocess doesn't need a catalog (e.g. if
+	 * we remove test_decoding), we should separate out tables for the follow
+	 * subprocess into their own database.
+	 */
+	if (!copydb_fetch_schema_and_prepare_specs(copySpecs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
 	/*
 	 * Preparation and snapshot are now done, time to fork our two main worker
 	 * processes.
