@@ -144,7 +144,6 @@ cli_pprint_json(JSON_Value *js)
 
 	/* free intermediate memory */
 	json_free_serialized_string(serialized_string);
-	json_value_free(js);
 }
 
 
@@ -454,6 +453,60 @@ cli_copydb_getenv(CopyDBOptions *options)
 		}
 	}
 
+	/*
+	 * When --skip-db-properties has not been used, check
+	 * PGCOPYDB_SKIP_DB_PROPERTIES
+	 */
+	if (!options->skipDBproperties)
+	{
+		if (env_exists(PGCOPYDB_SKIP_DB_PROPERTIES))
+		{
+			char SKIP_DB_PROPS[BUFSIZE] = { 0 };
+
+			if (!get_env_copy(PGCOPYDB_SKIP_DB_PROPERTIES,
+							  SKIP_DB_PROPS,
+							  sizeof(SKIP_DB_PROPS)))
+			{
+				/* errors have already been logged */
+				++errors;
+			}
+			else if (!parse_bool(SKIP_DB_PROPS, &(options->skipDBproperties)))
+			{
+				log_error("Failed to parse environment variable \"%s\" "
+						  "value \"%s\", expected a boolean (on/off)",
+						  PGCOPYDB_SKIP_DB_PROPERTIES,
+						  SKIP_DB_PROPS);
+				++errors;
+			}
+		}
+	}
+
+	/* when --no-tablespaces has not been used, check PGCOPYDB_SKIP_TABLESPACES */
+	if (!options->restoreOptions.noTableSpaces)
+	{
+		if (env_exists(PGCOPYDB_SKIP_TABLESPACES))
+		{
+			char SKIP_TABLESPACES[BUFSIZE] = { 0 };
+
+			if (!get_env_copy(PGCOPYDB_SKIP_TABLESPACES,
+							  SKIP_TABLESPACES,
+							  sizeof(SKIP_TABLESPACES)))
+			{
+				/* errors have already been logged */
+				++errors;
+			}
+			else if (!parse_bool(SKIP_TABLESPACES,
+								 &(options->restoreOptions.noTableSpaces)))
+			{
+				log_error("Failed to parse environment variable \"%s\" "
+						  "value \"%s\", expected a boolean (on/off)",
+						  PGCOPYDB_SKIP_TABLESPACES,
+						  SKIP_TABLESPACES);
+				++errors;
+			}
+		}
+	}
+
 	return errors == 0;
 }
 
@@ -692,31 +745,33 @@ cli_read_one_line(const char *filename,
 	}
 
 	/* make sure to use only the first line of the file, without \n */
-	char *lines[BUFSIZE] = { 0 };
-	int lineCount = splitLines(contents, lines, BUFSIZE);
+	LinesBuffer lbuf = { 0 };
 
-	if (lineCount != 1)
+	if (!splitLines(&lbuf, contents))
 	{
-		log_error("Failed to parse %s file \"%s\"", name, filename);
-		free(contents);
+		/* errors have already been logged */
 		return false;
 	}
 
-	if (size < (strlen(lines[0]) + 1))
+	if (lbuf.count != 1)
+	{
+		log_error("Failed to parse %s file \"%s\"", name, filename);
+		return false;
+	}
+
+	if (size < (strlen(lbuf.lines[0]) + 1))
 	{
 		log_error("Failed to parse %s \"%s\" with %lld bytes, "
 				  "pgcopydb supports only snapshot references up to %lld bytes",
 				  name,
-				  lines[0],
-				  (long long) strlen(lines[0]) + 1,
+				  lbuf.lines[0],
+				  (long long) strlen(lbuf.lines[0]) + 1,
 				  (long long) size);
-		free(contents);
 		return false;
 	}
 
 	/* publish the one line to the snapshot variable */
-	strlcpy(target, lines[0], size);
-	free(contents);
+	strlcpy(target, lbuf.lines[0], size);
 
 	return true;
 }
@@ -756,6 +811,8 @@ cli_copy_db_getopts(int argc, char **argv)
 		{ "skip-ext-comments", no_argument, NULL, 'M' },
 		{ "skip-collations", no_argument, NULL, 'l' },
 		{ "skip-vacuum", no_argument, NULL, 'U' },
+		{ "skip-db-properties", no_argument, NULL, 'g' },
+		{ "no-tablespaces", no_argument, NULL, 'y' },
 		{ "filter", required_argument, NULL, 'F' },
 		{ "filters", required_argument, NULL, 'F' },
 		{ "requirements", required_argument, NULL, 'Q' },
@@ -797,9 +854,11 @@ cli_copy_db_getopts(int argc, char **argv)
 		exit(EXIT_CODE_BAD_ARGS);
 	}
 
+	const char *optstring =
+		"S:T:D:J:I:b:L:cAPOXj:xBeMlUgyF:F:Q:irRCN:fp:ws:o:tE:Vvdzqh";
+
 	while ((c = getopt_long(argc, argv,
-							"S:T:D:J:I:b:L:cAPOXj:xBeMlUF:F:Q:irRCN:fp:ws:o:tE:Vvdzqh",
-							long_options, &option_index)) != -1)
+							optstring, long_options, &option_index)) != -1)
 	{
 		switch (c)
 		{
@@ -999,6 +1058,13 @@ cli_copy_db_getopts(int argc, char **argv)
 				break;
 			}
 
+			case 'g':
+			{
+				options.skipDBproperties = true;
+				log_trace("--skip-db-properties");
+				break;
+			}
+
 			case 'i':
 			{
 				options.failFast = true;
@@ -1181,6 +1247,13 @@ cli_copy_db_getopts(int argc, char **argv)
 				break;
 			}
 
+			case 'y':
+			{
+				options.restoreOptions.noTableSpaces = true;
+				log_trace("--no-tablespaces");
+				break;
+			}
+
 			case '?':
 			default:
 			{
