@@ -45,6 +45,7 @@ static CommandLine copy_db_command =
 		"  --no-owner            Do not set ownership of objects to match the original database\n"
 		"  --no-acl              Prevent restoration of access privileges (grant/revoke commands).\n"
 		"  --no-comments         Do not output commands to restore comments\n"
+		"  --no-tablespaces      Do not output commands to select tablespaces\n"
 		"  --skip-large-objects  Skip copying large objects (blobs)\n"
 		"  --filters <filename>  Use the filters defined in <filename>\n"
 		"  --fail-fast           Abort early in case of error\n"
@@ -242,6 +243,14 @@ cli_copy_schema(int argc, char **argv)
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
+	/* fetch schema information from source catalogs, including filtering */
+	if (!copydb_fetch_schema_and_prepare_specs(&copySpecs))
+	{
+		/* errors have already been logged */
+		(void) copydb_close_snapshot(&copySpecs);
+		exit(EXIT_CODE_TARGET);
+	}
+
 	if (!copydb_dump_source_schema(&copySpecs,
 								   copySpecs.sourceSnapshot.snapshot,
 								   PG_DUMP_SECTION_SCHEMA))
@@ -251,14 +260,6 @@ cli_copy_schema(int argc, char **argv)
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	/* fetch schema information from source catalogs, including filtering */
-	if (!copydb_fetch_schema_and_prepare_specs(&copySpecs))
-	{
-		/* errors have already been logged */
-		(void) copydb_close_snapshot(&copySpecs);
-		exit(EXIT_CODE_TARGET);
-	}
-
 	/* now close the snapshot we kept for the whole operation */
 	(void) copydb_close_snapshot(&copySpecs);
 
@@ -413,7 +414,9 @@ cli_copy_sequences(int argc, char **argv)
 		exit(EXIT_CODE_TARGET);
 	}
 
-	if (!copydb_copy_all_sequences(&copySpecs))
+	bool reset = true;
+
+	if (!copydb_copy_all_sequences(&copySpecs, reset))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
@@ -629,18 +632,28 @@ cli_copy_extensions(int argc, char **argv)
 	if (!copydb_fetch_schema_and_prepare_specs(&copySpecs))
 	{
 		/* errors have already been logged */
-		(void) copydb_close_snapshot(&copySpecs);
-		exit(EXIT_CODE_TARGET);
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	/* time to close the catalogs now */
+	if (!catalog_close_from_specs(&copySpecs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
 	bool createExtensions = true;
 
-	if (!copydb_copy_extensions(&copySpecs, createExtensions))
+	if (!copydb_start_extension_data_process(&copySpecs, createExtensions))
 	{
 		/* errors have already been logged */
-		exit(EXIT_CODE_TARGET);
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	/* now close the snapshot we kept for the whole operation */
-	(void) copydb_close_snapshot(&copySpecs);
+	if (!copydb_wait_for_subprocesses(copySpecs.failFast))
+	{
+		log_error("Some sub-processes have exited with error status, "
+				  "see above for details");
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
 }
