@@ -44,6 +44,7 @@ static CommandLine restore_schema_command =
 		"  --no-owner           Do not set ownership of objects to match the original database\n"
 		"  --no-acl             Prevent restoration of access privileges (grant/revoke commands).\n"
 		"  --no-comments        Do not output commands to restore comments\n"
+		"  --no-tablespaces     Do not output commands to select tablespaces\n"
 		"  --filters <filename> Use the filters defined in <filename>\n"
 		"  --restart            Allow restarting when temp files exist already\n"
 		"  --resume             Allow resuming operations after a failure\n"
@@ -64,6 +65,7 @@ static CommandLine restore_schema_pre_data_command =
 		"  --no-owner           Do not set ownership of objects to match the original database\n"
 		"  --no-acl             Prevent restoration of access privileges (grant/revoke commands).\n"
 		"  --no-comments        Do not output commands to restore comments\n"
+		"  --no-tablespaces     Do not output commands to select tablespaces\n"
 		"  --skip-extensions    Skip restoring extensions\n" \
 		"  --skip-ext-comments  Skip restoring COMMENT ON EXTENSION\n" \
 		"  --filters <filename> Use the filters defined in <filename>\n"
@@ -85,6 +87,7 @@ static CommandLine restore_schema_post_data_command =
 		"  --no-owner           Do not set ownership of objects to match the original database\n"
 		"  --no-acl             Prevent restoration of access privileges (grant/revoke commands).\n"
 		"  --no-comments        Do not output commands to restore comments\n"
+		"  --no-tablespaces     Do not output commands to select tablespaces\n"
 		"  --skip-extensions    Skip restoring extensions\n" \
 		"  --skip-ext-comments  Skip restoring COMMENT ON EXTENSION\n" \
 		"  --filters <filename> Use the filters defined in <filename>\n"
@@ -585,8 +588,6 @@ cli_restore_schema_parse_list(int argc, char **argv)
 					item->restoreListName ? item->restoreListName : "");
 		}
 
-		FreeArchiveContentArray(&contents);
-
 		exit(EXIT_CODE_QUIT);
 	}
 
