@@ -224,6 +224,7 @@ typedef struct CopyDataSpec
 	bool skipCommentOnExtension;
 	bool skipCollations;
 	bool skipVacuum;
+	bool skipDBproperties;
 	bool noRolesPasswords;
 
 	bool restart;
@@ -306,8 +307,6 @@ bool copydb_init_table_specs(CopyTableDataSpec *tableSpecs,
 							 SourceTable *source,
 							 int partNumber);
 
-void FreeCopyTableDataSpec(CopyTableDataSpec *tableSpecs);
-
 bool copydb_export_snapshot(TransactionSnapshot *snapshot);
 
 bool copydb_fatal_exit(void);
@@ -326,6 +325,7 @@ bool catalog_init_from_specs(CopyDataSpec *copySpecs);
 bool catalog_open_from_specs(CopyDataSpec *copySpecs);
 bool catalog_close_from_specs(CopyDataSpec *copySpecs);
 bool catalog_register_setup_from_specs(CopyDataSpec *copySpecs);
+bool catalog_update_setup(CopyDataSpec *copySpecs);
 
 /* snapshot.c */
 bool copydb_copy_snapshot(CopyDataSpec *specs, TransactionSnapshot *snapshot);
@@ -342,7 +342,8 @@ bool snapshot_write_slot(const char *filename, ReplicationSlot *slot);
 bool snapshot_read_slot(const char *filename, ReplicationSlot *slot);
 
 /* extensions.c */
-bool copydb_start_extension_data_process(CopyDataSpec *specs);
+bool copydb_start_extension_data_process(CopyDataSpec *specs,
+										 bool createExtensions);
 bool copydb_copy_extensions(CopyDataSpec *copySpecs, bool createExtensions);
 
 bool copydb_parse_extensions_requirements(CopyDataSpec *copySpecs,
@@ -395,14 +396,14 @@ bool copydb_target_drop_tables(CopyDataSpec *specs);
 bool copydb_target_finalize_schema(CopyDataSpec *specs);
 
 bool copydb_objectid_has_been_processed_already(CopyDataSpec *specs,
-												uint32_t oid);
+												ArchiveContentItem *item);
 
 bool copydb_write_restore_list(CopyDataSpec *specs, PostgresDumpSection section);
 
 /* sequences.c */
-bool copydb_copy_all_sequences(CopyDataSpec *specs);
+bool copydb_copy_all_sequences(CopyDataSpec *specs, bool reset);
 bool copydb_start_seq_process(CopyDataSpec *specs);
-bool copydb_prepare_sequence_specs(CopyDataSpec *specs, PGSQL *pgsql);
+bool copydb_prepare_sequence_specs(CopyDataSpec *specs, PGSQL *pgsql, bool reset);
 
 /* copydb_schema.c */
 bool copydb_fetch_schema_and_prepare_specs(CopyDataSpec *specs);
@@ -412,11 +413,12 @@ bool copydb_objectid_is_filtered_out(CopyDataSpec *specs,
 
 bool copydb_prepare_table_specs(CopyDataSpec *specs, PGSQL *pgsql);
 bool copydb_prepare_index_specs(CopyDataSpec *specs, PGSQL *pgsql);
+bool copydb_prepare_namespace_specs(CopyDataSpec *specs, PGSQL *pgsql);
 bool copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql);
 
 bool copydb_prepare_target_catalog(CopyDataSpec *specs);
 bool copydb_schema_already_exists(CopyDataSpec *specs,
-								  const char *restoreListName,
+								  uint32_t sourceOid,
 								  bool *exists);
 
 /* table-data.c */
@@ -445,6 +447,7 @@ bool copydb_copy_table(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 
 bool copydb_table_create_lockfile(CopyDataSpec *specs,
 								  CopyTableDataSpec *tableSpecs,
+								  PGSQL *dst,
 								  bool *isDone);
 
 bool copydb_mark_table_as_done(CopyDataSpec *specs,
@@ -457,11 +460,9 @@ bool copydb_table_parts_are_all_done(CopyDataSpec *specs,
 
 bool copydb_prepare_copy_query(CopyTableDataSpec *tableSpecs, CopyArgs *args);
 
-bool copydb_prepare_copy_query_attrlist(CopyTableDataSpec *tableSpecs,
-										PQExpBuffer attrList);
-
 bool copydb_prepare_summary_command(CopyTableDataSpec *tableSpecs);
 
+bool copydb_check_table_exists(PGSQL *pgsql, SourceTable *table, bool *exists);
 
 /* blobs.c */
 bool copydb_start_blob_process(CopyDataSpec *specs);
@@ -517,7 +518,6 @@ bool print_summary(CopyDataSpec *specs);
 bool summary_prepare_toplevel_durations(CopyDataSpec *specs);
 bool prepare_summary_table(Summary *summary, CopyDataSpec *specs);
 
-bool summary_lookup_oid(DatabaseCatalog *catalog, uint32_t oid, bool *done);
 bool summary_oid_done_fetch(SQLiteQuery *query);
 
 /*
