@@ -180,7 +180,7 @@ copydb_fetch_schema_and_prepare_specs(CopyDataSpec *specs)
 
 /*
  * copydb_fetch_source_catalog_setup initializes our local catalog cache and
- * checks th setup and cache state.
+ * checks the setup and cache state.
  */
 static bool
 copydb_fetch_source_catalog_setup(CopyDataSpec *specs)
@@ -232,6 +232,7 @@ copydb_fetch_source_catalog_setup(CopyDataSpec *specs)
 
 		/* compute "allDone" in the context of a sourceDB */
 		if (s->section == DATA_SECTION_DATABASE_PROPERTIES ||
+			s->section == DATA_SECTION_NAMESPACES ||
 			s->section == DATA_SECTION_TABLE_DATA ||
 			s->section == DATA_SECTION_SET_SEQUENCES ||
 			s->section == DATA_SECTION_INDEXES ||
@@ -408,6 +409,40 @@ copydb_fetch_previous_run_state(CopyDataSpec *specs)
 }
 
 
+/*
+ * Retrieve a list of namespaces from the source databases and store them
+ * in the source catalog and register the section as fetched.
+ */
+bool
+copydb_prepare_namespace_specs(CopyDataSpec *specs, PGSQL *pgsql)
+{
+	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
+
+	TopLevelTiming timing = {
+		.label = CopyDataSectionToString(DATA_SECTION_NAMESPACES)
+	};
+
+	(void) catalog_start_timing(&timing);
+
+	if (!schema_list_schemas(pgsql, sourceDB))
+	{
+		log_error("Failed to prepare namespace specs in our catalogs, "
+				  "see above for details");
+		return false;
+	}
+
+	(void) catalog_stop_timing(&timing);
+
+	if (!catalog_register_section(sourceDB, &timing))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	return true;
+}
+
+
 /*
  * copydb_fetch_source_schema is a utility function for the previous definition
  * copydb_fetch_schema_and_prepare_specs.
@@ -464,13 +499,31 @@ copydb_fetch_source_schema(CopyDataSpec *specs, PGSQL *src)
 				 "on the source database discards pg_table_size() caching");
 	}
 
+	if (!semaphore_lock(&(sourceDB->sema)))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	if (!catalog_begin(sourceDB, false))
+	{
+		/* errors have already been logged */
+		(void) semaphore_unlock(&(sourceDB->sema));
+		return false;
+	}
+
 	/*
 	 * Grab the source database properties to be able to install them again on
 	 * the target, using ALTER DATABASE SET or ALTER USER IN DATABASE SET.
 	 */
-	if ((specs->section == DATA_SECTION_ALL ||
-		 specs->section == DATA_SECTION_DATABASE_PROPERTIES) &&
-		!sourceDB->sections[DATA_SECTION_DATABASE_PROPERTIES].fetched)
+	if (specs->skipDBproperties)
+	{
+		log_notice("Skipping ALTER DATABASE SET operations, "
+				   "see --skip-db-properties");
+	}
+	else if ((specs->section == DATA_SECTION_ALL ||
+			  specs->section == DATA_SECTION_DATABASE_PROPERTIES) &&
+			 !sourceDB->sections[DATA_SECTION_DATABASE_PROPERTIES].fetched)
 	{
 		TopLevelTiming timing = {
 			.label = CopyDataSectionToString(DATA_SECTION_DATABASE_PROPERTIES)
@@ -494,16 +547,6 @@ copydb_fetch_source_schema(CopyDataSpec *specs, PGSQL *src)
 		}
 	}
 
-	/*
-	 * First, if it doesn't exist yet, create the pgcopydb.table_size table.
-	 * Keep track of whether we had to create that table, if we did, it is
-	 * expected that we DROP it before the end of this transaction.
-	 *
-	 * In order to allow for users to prepare that table in advance, we do not
-	 * use a TEMP table here.
-	 */
-	bool createdTableSizeTable = false;
-
 	/* now fetch the list of tables from the source database */
 	if ((specs->section == DATA_SECTION_ALL ||
 		 specs->section == DATA_SECTION_TABLE_DATA ||
@@ -513,18 +556,17 @@ copydb_fetch_source_schema(CopyDataSpec *specs, PGSQL *src)
 		/* copydb_fetch_filtered_oids() needs the table size table around */
 		if (!schema_prepare_pgcopydb_table_size(src,
 												&(specs->filters),
-												specs->hasDBCreatePrivilege,
-												false, /* cache */
-												false, /* dropCache */
-												&createdTableSizeTable))
+												sourceDB))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(sourceDB->sema));
 			return false;
 		}
 
 		if (!copydb_prepare_table_specs(specs, src))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(sourceDB->sema));
 			return false;
 		}
 	}
@@ -538,6 +580,7 @@ copydb_fetch_source_schema(CopyDataSpec *specs, PGSQL *src)
 		if (!copydb_prepare_index_specs(specs, src))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(sourceDB->sema));
 			return false;
 		}
 	}
@@ -546,32 +589,56 @@ copydb_fetch_source_schema(CopyDataSpec *specs, PGSQL *src)
 		 specs->section == DATA_SECTION_SET_SEQUENCES) &&
 		!sourceDB->sections[DATA_SECTION_SET_SEQUENCES].fetched)
 	{
-		if (!copydb_prepare_sequence_specs(specs, src))
+		bool reset = false;
+
+		if (!copydb_prepare_sequence_specs(specs, src, reset))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(sourceDB->sema));
 			return false;
 		}
 	}
 
+	if ((specs->section == DATA_SECTION_ALL ||
+		 specs->section == DATA_SECTION_NAMESPACES) &&
+		!sourceDB->sections[DATA_SECTION_NAMESPACES].fetched)
+	{
+		if (!copydb_prepare_namespace_specs(specs, src))
+		{
+			/* errors have already been logged */
+			(void) semaphore_unlock(&(sourceDB->sema));
+			return false;
+		}
+	}
+
+	/* now update --split-tables-larger-than and target pguri */
+	if (!catalog_update_setup(specs))
+	{
+		/* errors have already been logged */
+		(void) semaphore_unlock(&(sourceDB->sema));
+		return false;
+	}
+
 	/* prepare the Oids of objects that are filtered out */
 	if (specs->fetchFilteredOids)
 	{
 		if (!copydb_fetch_filtered_oids(specs, src))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(sourceDB->sema));
 			return false;
 		}
 	}
 
-	if (createdTableSizeTable)
+	if (!catalog_commit(sourceDB))
 	{
-		if (!schema_drop_pgcopydb_table_size(src))
-		{
-			/* errors have already been logged */
-			return false;
-		}
+		/* errors have already been logged */
+		(void) semaphore_unlock(&(sourceDB->sema));
+		return false;
 	}
 
+	(void) semaphore_unlock(&(sourceDB->sema));
+
 	return true;
 }
 
@@ -732,6 +799,14 @@ copydb_prepare_table_specs_hook(void *ctx, SourceTable *source)
 					  source->qname);
 			return false;
 		}
+
+		/* fetch the relpages for the table after ANALYZE */
+		if (!schema_list_relpages(context->pgsql, source, sourceDB))
+		{
+			log_error("Failed to fetch table %s relpages",
+					  source->qname);
+			return false;
+		}
 	}
 	else if (!streq(source->amname, "heap"))
 	{
@@ -895,6 +970,12 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 
 	CatalogCounts count = { 0 };
 
+	if (!semaphore_lock(&(filtersDB->sema)))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	/* now, are we doing extensions? */
 	if ((specs->section == DATA_SECTION_ALL ||
 		 specs->section == DATA_SECTION_EXTENSIONS) &&
@@ -906,9 +987,19 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 
 		(void) catalog_start_timing(&timing);
 
+		/* fetch the list of schemas that extensions depend on */
+		if (!schema_list_ext_schemas(pgsql, filtersDB))
+		{
+			/* errors have already been logged */
+			(void) semaphore_unlock(&(filtersDB->sema));
+			return false;
+		}
+
+		/* and fetch the list of extensions we want to skip */
 		if (!schema_list_extensions(pgsql, filtersDB))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
@@ -917,12 +1008,14 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		if (!catalog_register_section(filtersDB, &timing))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
 		if (!catalog_count_objects(filtersDB, &count))
 		{
 			log_error("Failed to count objects in our catalogs");
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
@@ -930,17 +1023,6 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 				 (long long) count.extensions);
 	}
 
-	if (specs->skipExtensions &&
-		!filtersDB->sections[DATA_SECTION_EXTENSIONS].fetched)
-	{
-		/* fetch the list of schemas that extensions depend on */
-		if (!schema_list_ext_schemas(pgsql, filtersDB))
-		{
-			/* errors have already been logged */
-			return false;
-		}
-	}
-
 	if (specs->skipCollations &&
 		!filtersDB->sections[DATA_SECTION_COLLATIONS].fetched)
 	{
@@ -953,6 +1035,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		if (!schema_list_collations(pgsql, filtersDB))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
@@ -961,12 +1044,14 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		if (!catalog_register_section(filtersDB, &timing))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
 		if (!catalog_count_objects(filtersDB, &count))
 		{
 			log_error("Failed to count indexes and constraints in our catalogs");
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
@@ -992,6 +1077,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 			if (!catalog_attach(filtersDB, sourceDB, "source"))
 			{
 				/* errors have already been logged */
+				(void) semaphore_unlock(&(filtersDB->sema));
 				return false;
 			}
 
@@ -1001,10 +1087,13 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 
 			(void) catalog_start_timing(&timing);
 
-			if (!catalog_prepare_filter(filtersDB))
+			if (!catalog_prepare_filter(filtersDB,
+										specs->skipExtensions,
+										specs->skipCollations))
 			{
 				log_error("Failed to prepare filtering hash-table, "
 						  "see above for details");
+				(void) semaphore_unlock(&(filtersDB->sema));
 				return false;
 			}
 
@@ -1013,10 +1102,13 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 			if (!catalog_register_section(filtersDB, &timing))
 			{
 				/* errors have already been logged */
+				(void) semaphore_unlock(&(filtersDB->sema));
 				return false;
 			}
 		}
 
+		(void) semaphore_unlock(&(filtersDB->sema));
+
 		return true;
 	}
 
@@ -1037,6 +1129,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		{
 			/* errors have already been logged */
 			filters->type = type;
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
@@ -1046,6 +1139,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		{
 			/* errors have already been logged */
 			filters->type = type;
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 	}
@@ -1065,6 +1159,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		{
 			/* errors have already been logged */
 			filters->type = type;
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
@@ -1074,6 +1169,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		{
 			/* errors have already been logged */
 			filters->type = type;
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
@@ -1087,6 +1183,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		if (!catalog_register_section(filtersDB, &cTiming))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 	}
@@ -1105,6 +1202,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		{
 			/* errors have already been logged */
 			filters->type = type;
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
@@ -1114,6 +1212,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		{
 			/* errors have already been logged */
 			filters->type = type;
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 	}
@@ -1130,6 +1229,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		{
 			/* errors have already been logged */
 			filters->type = type;
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
@@ -1139,6 +1239,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		{
 			/* errors have already been logged */
 			filters->type = type;
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 	}
@@ -1152,6 +1253,7 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 	if (!catalog_attach(filtersDB, sourceDB, "source"))
 	{
 		/* errors have already been logged */
+		(void) semaphore_unlock(&(filtersDB->sema));
 		return false;
 	}
 
@@ -1165,10 +1267,13 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 
 		(void) catalog_start_timing(&timing);
 
-		if (!catalog_prepare_filter(filtersDB))
+		if (!catalog_prepare_filter(filtersDB,
+									specs->skipExtensions,
+									specs->skipCollations))
 		{
 			log_error("Failed to prepare filtering hash-table, "
 					  "see above for details");
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 
@@ -1177,10 +1282,13 @@ copydb_fetch_filtered_oids(CopyDataSpec *specs, PGSQL *pgsql)
 		if (!catalog_register_section(filtersDB, &timing))
 		{
 			/* errors have already been logged */
+			(void) semaphore_unlock(&(filtersDB->sema));
 			return false;
 		}
 	}
 
+	(void) semaphore_unlock(&(filtersDB->sema));
+
 	return true;
 }
 
@@ -1213,23 +1321,39 @@ copydb_prepare_target_catalog(CopyDataSpec *specs)
 	 */
 	DatabaseCatalog *targetDB = &(specs->catalogs.target);
 
+	if (!semaphore_lock(&(targetDB->sema)))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	if (!catalog_drop_schema(targetDB) ||
 		!catalog_create_schema(targetDB))
 	{
 		log_error("Failed to clean-up the target catalog cache, "
 				  "see above for details");
+		(void) semaphore_unlock(&(targetDB->sema));
 		return false;
 	}
 
 	if (!pgsql_init(&dst, specs->connStrings.target_pguri, PGSQL_CONN_TARGET))
 	{
 		/* errors have already been logged */
+		(void) semaphore_unlock(&(targetDB->sema));
 		return false;
 	}
 
 	if (!pgsql_begin(&dst))
 	{
 		/* errors have already been logged */
+		(void) semaphore_unlock(&(targetDB->sema));
+		return false;
+	}
+
+	if (!catalog_begin(targetDB, false))
+	{
+		/* errors have already been logged */
+		(void) semaphore_unlock(&(targetDB->sema));
 		return false;
 	}
 
@@ -1242,6 +1366,7 @@ copydb_prepare_target_catalog(CopyDataSpec *specs)
 	if (!schema_list_schemas(&dst, targetDB))
 	{
 		log_error("Failed to list schemas on the target database");
+		(void) semaphore_unlock(&(targetDB->sema));
 		return false;
 	}
 
@@ -1256,6 +1381,7 @@ copydb_prepare_target_catalog(CopyDataSpec *specs)
 	if (!schema_list_roles(&dst, targetDB))
 	{
 		log_error("Failed to list roles on the target database");
+		(void) semaphore_unlock(&(targetDB->sema));
 		return false;
 	}
 
@@ -1274,18 +1400,28 @@ copydb_prepare_target_catalog(CopyDataSpec *specs)
 		log_error("Failed to DELETE all target catalog indexes "
 				  "in our internal catalogs (cache invalidation), "
 				  "see above for details");
+		(void) semaphore_unlock(&(targetDB->sema));
 		return false;
 	}
 
 	if (!schema_list_all_indexes(&dst, &targetDBfilter, targetDB))
 	{
 		/* errors have already been logged */
+		(void) semaphore_unlock(&(targetDB->sema));
+		return false;
+	}
+
+	if (!catalog_commit(targetDB))
+	{
+		/* errors have already been logged */
+		(void) semaphore_unlock(&(targetDB->sema));
 		return false;
 	}
 
 	if (!pgsql_commit(&dst))
 	{
 		/* errors have already been logged */
+		(void) semaphore_unlock(&(targetDB->sema));
 		return false;
 	}
 
@@ -1294,6 +1430,7 @@ copydb_prepare_target_catalog(CopyDataSpec *specs)
 	if (!catalog_count_objects(targetDB, &count))
 	{
 		log_error("Failed to count indexes and constraints in our catalogs");
+		(void) semaphore_unlock(&(targetDB->sema));
 		return false;
 	}
 
@@ -1302,6 +1439,8 @@ copydb_prepare_target_catalog(CopyDataSpec *specs)
 			 (long long) count.indexes,
 			 (long long) count.constraints);
 
+	(void) semaphore_unlock(&(targetDB->sema));
+
 	return true;
 }
 
@@ -1312,21 +1451,39 @@ copydb_prepare_target_catalog(CopyDataSpec *specs)
  */
 bool
 copydb_schema_already_exists(CopyDataSpec *specs,
-							 const char *restoreListName,
+							 uint32_t sourceOid,
 							 bool *exists)
 {
-	DatabaseCatalog *targetDB = &(specs->catalogs.target);
-	SourceSchema schema = { 0 };
+	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
+	SourceSchema sourceResults = { 0 };
 
-	if (!catalog_lookup_s_namespace_by_rlname(targetDB,
-											  restoreListName,
-											  &schema))
+	if (!catalog_lookup_s_namespace_by_oid(sourceDB,
+										   sourceOid,
+										   &sourceResults))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
-	*exists = (schema.oid != 0);
+	if (sourceResults.oid == 0)
+	{
+		log_error("Failed to find schema with oid %d in source database",
+				  sourceOid);
+		return false;
+	}
+
+	DatabaseCatalog *targetDB = &(specs->catalogs.target);
+	SourceSchema targetResults = { 0 };
+
+	if (!catalog_lookup_s_namespace_by_nspname(targetDB,
+											   sourceResults.nspname,
+											   &targetResults))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	*exists = (targetResults.oid != 0);
 
 	return true;
 }
