@@ -31,23 +31,58 @@ static bool copydb_copy_database_properties_hook(void *ctx,
 
 
 /*
- * copydb_objectid_has_been_processed_already returns true when a doneFile
- * could be found on-disk for the given target object OID.
+ * copydb_objectid_has_been_processed_already returns true when the given
+ * target object OID is found in our SQLite summary catalogs. This only applies
+ * to indexes or constraints, as TABLE DATA is not part of either the
+ * --pre-data parts of the schema nor the --post-data parts of the schema.
  */
 bool
-copydb_objectid_has_been_processed_already(CopyDataSpec *specs, uint32_t oid)
+copydb_objectid_has_been_processed_already(CopyDataSpec *specs,
+										   ArchiveContentItem *item)
 {
 	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
 
-	bool done = false;
+	uint32_t oid = item->objectOid;
 
-	if (!summary_lookup_oid(sourceDB, oid, &done))
+	switch (item->desc)
 	{
-		/* errors have aleady been logged */
-		return false;
+		case ARCHIVE_TAG_INDEX:
+		{
+			SourceIndex index = { .indexOid = oid };
+			CopyIndexSpec indexSpecs = { .sourceIndex = &index };
+
+			if (!summary_lookup_index(sourceDB, &indexSpecs))
+			{
+				/* errors have aleady been logged */
+				return false;
+			}
+
+			return indexSpecs.summary.doneTime > 0;
+		}
+
+		case ARCHIVE_TAG_CONSTRAINT:
+		{
+			SourceIndex index = { .constraintOid = oid };
+			CopyIndexSpec indexSpecs = { .sourceIndex = &index };
+
+			if (!summary_lookup_constraint(sourceDB, &indexSpecs))
+			{
+				/* errors have aleady been logged */
+				return false;
+			}
+
+			return indexSpecs.summary.doneTime > 0;
+		}
+
+		/* we don't have internal pgcopydb support for other objects */
+		default:
+		{
+			return false;
+		}
 	}
 
-	return done;
+	/* keep compiler happy */
+	return false;
 }
 
 
@@ -62,6 +97,13 @@ copydb_dump_source_schema(CopyDataSpec *specs,
 {
 	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
 
+	if (specs->runState.schemaDumpIsDone)
+	{
+		log_info("Skipping pg_dump for pre-data/post-data section, "
+				 "done on a previous run");
+		return true;
+	}
+
 	if (!summary_start_timing(sourceDB, TIMING_SECTION_DUMP_SCHEMA))
 	{
 		/* errors have already been logged */
@@ -133,12 +175,6 @@ copydb_target_prepare_schema(CopyDataSpec *specs)
 {
 	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
 
-	if (!summary_start_timing(sourceDB, TIMING_SECTION_PREPARE_SCHEMA))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
 	if (!file_exists(specs->dumpPaths.preFilename))
 	{
 		log_fatal("File \"%s\" does not exists", specs->dumpPaths.preFilename);
@@ -152,6 +188,12 @@ copydb_target_prepare_schema(CopyDataSpec *specs)
 		return true;
 	}
 
+	if (!summary_start_timing(sourceDB, TIMING_SECTION_PREPARE_SCHEMA))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	/*
 	 * First restore the database properties (ALTER DATABASE SET).
 	 */
@@ -281,7 +323,14 @@ copydb_copy_database_properties_hook(void *ctx, SourceProperty *property)
 	PGSQL *dst = context->dst;
 	PGconn *conn = dst->connection;
 
-	const char *t_dbname = specs->connStrings.safeTargetPGURI.uriParams.dbname;
+	char *t_dbname = specs->connStrings.safeTargetPGURI.uriParams.dbname;
+	char *t_escaped_dbname = pgsql_escape_identifier(dst, t_dbname);
+
+	if (t_escaped_dbname == NULL)
+	{
+		/* errors are already logged */
+		return false;
+	}
 
 	/*
 	 * ALTER ROLE rolname IN DATABASE datname SET ...
@@ -301,7 +350,6 @@ copydb_copy_database_properties_hook(void *ctx, SourceProperty *property)
 		if (!catalog_lookup_s_role_by_name(targetDB, property->rolname, role))
 		{
 			/* errors have already been logged */
-			free(role);
 			return false;
 		}
 
@@ -312,7 +360,7 @@ copydb_copy_database_properties_hook(void *ctx, SourceProperty *property)
 			makeAlterConfigCommand(conn, property->setconfig,
 								   "ROLE", property->rolname,
 								   "DATABASE",
-								   t_dbname,
+								   t_escaped_dbname,
 								   command);
 
 			/* chomp the \n */
@@ -326,7 +374,6 @@ copydb_copy_database_properties_hook(void *ctx, SourceProperty *property)
 			if (!pgsql_execute(dst, command->data))
 			{
 				/* errors have already been logged */
-				free(role);
 				return false;
 			}
 
@@ -338,8 +385,6 @@ copydb_copy_database_properties_hook(void *ctx, SourceProperty *property)
 					 "does not exists on the target database",
 					 property->rolname);
 		}
-
-		free(role);
 	}
 
 	/*
@@ -347,10 +392,26 @@ copydb_copy_database_properties_hook(void *ctx, SourceProperty *property)
 	 */
 	else
 	{
+		bool exists = false;
+
+		if (!pgsql_configuration_exists(dst, property->setconfig, &exists))
+		{
+			/* errors have already been logged */
+			return false;
+		}
+
+		if (!exists)
+		{
+			log_warn("Skipping database property %s which "
+					 "does not exists on the target database",
+					 property->setconfig);
+			return true;
+		}
+
 		PQExpBuffer command = createPQExpBuffer();
 
 		makeAlterConfigCommand(conn, property->setconfig,
-							   "DATABASE", t_dbname,
+							   "DATABASE", t_escaped_dbname,
 							   NULL, NULL,
 							   command);
 
@@ -430,6 +491,8 @@ copydb_target_drop_tables(CopyDataSpec *specs)
 		return false;
 	}
 
+	log_notice("%s", query->data);
+
 	PGSQL dst = { 0 };
 
 	if (!pgsql_init(&dst, specs->connStrings.target_pguri, PGSQL_CONN_TARGET))
@@ -485,12 +548,6 @@ copydb_target_finalize_schema(CopyDataSpec *specs)
 {
 	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
 
-	if (!summary_start_timing(sourceDB, TIMING_SECTION_FINALIZE_SCHEMA))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
 	if (!file_exists(specs->dumpPaths.postFilename))
 	{
 		log_fatal("File \"%s\" does not exists", specs->dumpPaths.postFilename);
@@ -504,6 +561,12 @@ copydb_target_finalize_schema(CopyDataSpec *specs)
 		return true;
 	}
 
+	if (!summary_start_timing(sourceDB, TIMING_SECTION_FINALIZE_SCHEMA))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	if (!copydb_write_restore_list(specs, PG_DUMP_SECTION_POST_DATA))
 	{
 		log_error("Failed to prepare the pg_restore --use-list catalogs, "
@@ -595,7 +658,6 @@ copydb_write_restore_list(CopyDataSpec *specs, PostgresDumpSection section)
 						 &contents))
 	{
 		/* errors have already been logged */
-		FreeArchiveContentArray(&contents);
 		return false;
 	}
 
@@ -619,6 +681,16 @@ copydb_write_restore_list(CopyDataSpec *specs, PostgresDumpSection section)
 
 		bool skip = false;
 
+		/*
+		 * Always skip DATABASE objets as pgcopydb does not create the target
+		 * database.
+		 */
+		if (item->desc == ARCHIVE_TAG_DATABASE)
+		{
+			skip = true;
+			log_notice("Skipping DATABASE \"%s\"", name);
+		}
+
 		/*
 		 * Skip COMMENT ON EXTENSION when either of the option
 		 * --skip-extensions or --skip-ext-comment has been used.
@@ -637,7 +709,7 @@ copydb_write_restore_list(CopyDataSpec *specs, PostgresDumpSection section)
 		{
 			bool exists = false;
 
-			if (!copydb_schema_already_exists(specs, name, &exists))
+			if (!copydb_schema_already_exists(specs, oid, &exists))
 			{
 				log_error("Failed to check if restore name \"%s\" "
 						  "already exists",
@@ -658,7 +730,7 @@ copydb_write_restore_list(CopyDataSpec *specs, PostgresDumpSection section)
 			}
 		}
 
-		if (!skip && copydb_objectid_has_been_processed_already(specs, oid))
+		if (!skip && copydb_objectid_has_been_processed_already(specs, item))
 		{
 			skip = true;
 
@@ -710,7 +782,6 @@ copydb_write_restore_list(CopyDataSpec *specs, PostgresDumpSection section)
 	{
 		log_error("Failed to create pg_restore list file: out of memory");
 		destroyPQExpBuffer(listContents);
-		FreeArchiveContentArray(&contents);
 		return false;
 	}
 
@@ -720,12 +791,10 @@ copydb_write_restore_list(CopyDataSpec *specs, PostgresDumpSection section)
 	{
 		/* errors have already been logged */
 		destroyPQExpBuffer(listContents);
-		FreeArchiveContentArray(&contents);
 		return false;
 	}
 
 	destroyPQExpBuffer(listContents);
-	FreeArchiveContentArray(&contents);
 
 	return true;
 }
