@@ -296,6 +296,8 @@ copydb_index_worker(CopyDataSpec *specs)
 		}
 	}
 
+	pgsql_finish(&dst);
+
 	if (!catalog_delete_process(&(specs->catalogs.source), pid))
 	{
 		log_warn("Failed to delete catalog process entry for pid %d", pid);
@@ -304,12 +306,9 @@ copydb_index_worker(CopyDataSpec *specs)
 	if (!catalog_close_from_specs(specs))
 	{
 		/* errors have already been logged */
-		(void) pgsql_finish(&dst);
 		return false;
 	}
 
-	(void) pgsql_finish(&dst);
-
 	bool success = (stop == true && errors == 0);
 
 	if (errors > 0)
@@ -339,16 +338,12 @@ copydb_create_index_by_oid(CopyDataSpec *specs, PGSQL *dst, uint32_t indexOid)
 	if (!catalog_lookup_s_index(sourceDB, indexOid, index))
 	{
 		log_error("Failed to lookup index %u in our catalogs", indexOid);
-		free(index);
-		free(table);
 		return false;
 	}
 
 	if (!catalog_lookup_s_table(sourceDB, index->tableOid, 0, table))
 	{
 		log_error("Failed to lookup table %u in our catalogs", index->tableOid);
-		free(index);
-		free(table);
 		return false;
 	}
 
@@ -390,8 +385,6 @@ copydb_create_index_by_oid(CopyDataSpec *specs, PGSQL *dst, uint32_t indexOid)
 	if (!copydb_create_index(specs, dst, index, ifNotExists))
 	{
 		/* errors have already been logged */
-		free(index);
-		free(table);
 		return false;
 	}
 
@@ -410,8 +403,6 @@ copydb_create_index_by_oid(CopyDataSpec *specs, PGSQL *dst, uint32_t indexOid)
 									   &constraintsAreBeingBuilt))
 	{
 		/* errors have already been logged */
-		free(index);
-		free(table);
 		return false;
 	}
 
@@ -428,8 +419,6 @@ copydb_create_index_by_oid(CopyDataSpec *specs, PGSQL *dst, uint32_t indexOid)
 		{
 			log_error("Failed to create constraints for table %s",
 					  table->qname);
-			free(index);
-			free(table);
 			return false;
 		}
 
@@ -440,15 +429,11 @@ copydb_create_index_by_oid(CopyDataSpec *specs, PGSQL *dst, uint32_t indexOid)
 				log_error("Failed to queue VACUUM ANALYZE %s [%u]",
 						  table->qname,
 						  table->oid);
-				free(index);
-				free(table);
 				return false;
 			}
 		}
 	}
 
-	free(index);
-	free(table);
 
 	return true;
 }
@@ -526,13 +511,19 @@ copydb_table_indexes_are_done(CopyDataSpec *specs,
 }
 
 
+typedef struct IndexOIDArray
+{
+	int count;
+	uint32_t *array;            /* malloc'ed area */
+} IndexOIDArray;
+
 typedef struct QueueTableIndexesContext
 {
 	CopyDataSpec *specs;
 	CopyTableDataSpec *tableSpecs;
+	IndexOIDArray *indexArray;
 } QueueTableIndexesContext;
 
-
 /*
  * copydb_add_table_indexes sends a message to the CREATE INDEX process queue
  * to process indexes attached to the given table.
@@ -541,24 +532,83 @@ bool
 copydb_add_table_indexes(CopyDataSpec *specs, CopyTableDataSpec *tableSpecs)
 {
 	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
+	IndexOIDArray indexArray = { 0, NULL };
+
+	if (!catalog_s_table_count_indexes(sourceDB, tableSpecs->sourceTable))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	/*
+	 * While this COPY process is holding a catalog lock to iterate over the
+	 * indexes of the given table, CREATE INDEX processes are attempting to
+	 * grab the same catalog lock in order to fetch the metadata for the OID
+	 * received from the queue.
+	 *
+	 * To avoid grabbing the SQLite semaphore while doing the queue_send()
+	 * operation, we allocate an array of indexes OIDs in memory and fill the
+	 * information from the iterator callback function.
+	 *
+	 * That's important because the queue_send() operation contains a retry
+	 * loop in case of errors, one possible error is EGAIN (queue is full). We
+	 * need to make sure that the CREATE INDEX processes are able to empty the
+	 * queue (and thus grab the catalog lock) during the queue_send() retry
+	 * loop.
+	 */
+	int indexCount = tableSpecs->sourceTable->indexCount;
+	indexArray.count = 0;
+	indexArray.array = (uint32_t *) calloc(indexCount, sizeof(uint32_t));
+
+	if (indexArray.array == NULL)
+	{
+		log_error(ALLOCATION_FAILED_ERROR);
+		return false;
+	}
 
 	QueueTableIndexesContext context = {
 		.specs = specs,
-		.tableSpecs = tableSpecs
+		.tableSpecs = tableSpecs,
+		.indexArray = &indexArray
 	};
 
+	/* iterate over SQLite results and fill-in the indexArray */
 	if (!catalog_iter_s_index_table(sourceDB,
 									tableSpecs->sourceTable->nspname,
 									tableSpecs->sourceTable->relname,
 									&context,
 									&copydb_add_table_indexes_hook))
 	{
-		log_error("Failed to send table %s indexes to craete index queue, "
+		log_error("Failed to send table %s indexes to create index queue, "
 				  "see above for details",
 				  tableSpecs->sourceTable->qname);
 		return false;
 	}
 
+	/*
+	 * Now that we are no longer holding the catalog lock, walk through the
+	 * indexArray and send index OIDs to the create index queue, retrying the
+	 * operation if the queue is full.
+	 */
+	for (int i = 0; i < indexArray.count; i++)
+	{
+		QMessage mesg = {
+			.type = QMSG_TYPE_INDEXOID,
+			.data.oid = indexArray.array[i]
+		};
+
+		log_trace("Queueing index [%u] for table %s [%u]",
+				  mesg.data.oid,
+				  tableSpecs->sourceTable->qname,
+				  tableSpecs->sourceTable->oid);
+
+		if (!queue_send(&(specs->indexQueue), &mesg))
+		{
+			/* errors have already been logged */
+			return false;
+		}
+	}
+
 	return true;
 }
 
@@ -570,25 +620,9 @@ static bool
 copydb_add_table_indexes_hook(void *ctx, SourceIndex *index)
 {
 	QueueTableIndexesContext *context = (QueueTableIndexesContext *) ctx;
-	CopyDataSpec *specs = context->specs;
-	CopyTableDataSpec *tableSpecs = context->tableSpecs;
+	IndexOIDArray *indexArray = context->indexArray;
 
-	QMessage mesg = {
-		.type = QMSG_TYPE_INDEXOID,
-		.data.oid = index->indexOid
-	};
-
-	log_trace("Queueing index %s [%u] for table %s [%u]",
-			  index->indexQname,
-			  mesg.data.oid,
-			  tableSpecs->sourceTable->qname,
-			  tableSpecs->sourceTable->oid);
-
-	if (!queue_send(&(specs->indexQueue), &mesg))
-	{
-		/* errors have already been logged */
-		return false;
-	}
+	indexArray->array[(indexArray->count)++] = index->indexOid;
 
 	return true;
 }
@@ -731,16 +765,6 @@ copydb_create_index(CopyDataSpec *specs,
 	CopyIndexSpec indexSpecs = { .sourceIndex = index };
 	CopyIndexSummary *indexSummary = &(indexSpecs.summary);
 
-	/*
-	 * Prepare the summary CREATE INDEX command early so that we can insert it
-	 * in our SQLite catalogs.
-	 */
-	if (!copydb_prepare_create_index_command(&indexSpecs, ifNotExists))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
 	bool isConstraintIndex = index->constraintOid != 0;
 	bool skipCreateIndex = false;
 
@@ -782,6 +806,16 @@ copydb_create_index(CopyDataSpec *specs,
 
 	if (!skipCreateIndex)
 	{
+		/*
+		 * Prepare the CREATE INDEX command based on the index definition and
+		 * ifNotExists flag.
+		 */
+		if (!copydb_prepare_create_index_command(&indexSpecs, ifNotExists))
+		{
+			/* errors have already been logged */
+			return false;
+		}
+
 		log_notice("%s", indexSummary->command);
 
 		if (!pgsql_execute(dst, indexSummary->command))
@@ -1060,7 +1094,6 @@ copydb_create_constraints(CopyDataSpec *specs, PGSQL *dst, SourceTable *table)
 	{
 		log_error("Failed to count indexes for table %s in our target catalog",
 				  targetTable->qname);
-		free(targetTable);
 		return false;
 	}
 
@@ -1083,7 +1116,6 @@ copydb_create_constraints(CopyDataSpec *specs, PGSQL *dst, SourceTable *table)
 				  table->qname);
 	}
 
-	free(targetTable);
 
 	/*
 	 * Now iterate over the source database catalog list of indexes attached to
