@@ -35,7 +35,6 @@
 static bool updateStreamCounters(StreamContext *context,
 								 LogicalMessageMetadata *metadata);
 
-
 /*
  * stream_init_specs initializes Change Data Capture streaming specifications
  * from a copyDBSpecs structure.
@@ -543,6 +542,9 @@ startLogicalStreaming(StreamSpecs *specs)
 					 privateContext->counters.total > 0 ? "s" : "");
 		}
 
+		/* if we are going to retry, we need to rollback the last txn */
+		context.onRetry = retry;
+
 		/* sleep for one entire second before retrying */
 		if (retry)
 		{
@@ -622,7 +624,7 @@ streamCheckResumePosition(StreamSpecs *specs)
 				 LSN_FORMAT_ARGS(specs->endpos));
 	}
 
-	if (latestStreamedContent.count == 0)
+	if (latestStreamedContent.lbuf.count == 0)
 	{
 		if (specs->mode == STREAM_MODE_RECEIVE)
 		{
@@ -642,7 +644,7 @@ streamCheckResumePosition(StreamSpecs *specs)
 	else
 	{
 		/* lines are counted starting at zero */
-		int lastLineNb = latestStreamedContent.count - 1;
+		int lastLineNb = latestStreamedContent.lbuf.count - 1;
 
 		LogicalMessageMetadata *messages = latestStreamedContent.messages;
 		LogicalMessageMetadata *latest = &(messages[lastLineNb]);
@@ -654,7 +656,7 @@ streamCheckResumePosition(StreamSpecs *specs)
 				 LSN_FORMAT_ARGS(specs->startpos),
 				 latestStreamedContent.filename);
 
-		char *latestMessage = latestStreamedContent.lines[lastLineNb];
+		char *latestMessage = latestStreamedContent.lbuf.lines[lastLineNb];
 		log_notice("Resume replication from latest message: %s", latestMessage);
 	}
 
@@ -783,6 +785,32 @@ stream_write_json(LogicalStreamContext *context, bool previous)
 		return false;
 	}
 
+	if (context->onRetry)
+	{
+		/*
+		 * When retrying due to a transient network error or server conn
+		 * failure, we need to rollback the last incomplete transaction.
+		 *
+		 * Otherwise, we would end up with a partial transaction in the
+		 * JSON file, and the transform process would fail to process it.
+		 */
+		if (privateContext->transactionInProgress)
+		{
+			InternalMessage rollback = {
+				.action = STREAM_ACTION_ROLLBACK,
+				.lsn = context->cur_record_lsn
+			};
+
+			if (!stream_write_internal_message(context, &rollback))
+			{
+				/* errors have already been logged */
+				return false;
+			}
+		}
+
+		context->onRetry = false;
+	}
+
 	/* prepare a in-memory buffer with the whole data formatted in JSON */
 	PQExpBuffer buffer = createPQExpBuffer();
 
@@ -861,7 +889,6 @@ stream_write_json(LogicalStreamContext *context, bool previous)
 	}
 
 	destroyPQExpBuffer(buffer);
-	free(metadata->jsonBuffer);
 
 	/*
 	 * Maintain the transaction progress based on the BEGIN and COMMIT messages
@@ -1376,12 +1403,6 @@ streamFlush(LogicalStreamContext *context)
 {
 	StreamContext *privateContext = (StreamContext *) context->private;
 
-	/* when there is currently no file opened, just skip the flush operation */
-	if (privateContext->jsonFile == NULL)
-	{
-		return true;
-	}
-
 	log_debug("streamFlush: %X/%X %X/%X",
 			  LSN_FORMAT_ARGS(context->tracking->written_lsn),
 			  LSN_FORMAT_ARGS(context->cur_record_lsn));
@@ -1401,6 +1422,10 @@ streamFlush(LogicalStreamContext *context)
 			return false;
 		}
 
+		/*
+		 * streamKeepalive ensures we have a valid jsonFile by calling
+		 * streamRotateFile, so we can safely call fsync here.
+		 */
 		int fd = fileno(privateContext->jsonFile);
 
 		if (fsync(fd) != 0)
@@ -1931,25 +1956,31 @@ bool
 stream_read_file(StreamContent *content)
 {
 	long size = 0L;
+	char *contents = NULL;
 
-	if (!read_file(content->filename, &(content->buffer), &size))
+	if (!read_file(content->filename, &contents, &size))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
-	content->count = countLines(content->buffer);
-	content->lines = (char **) calloc(content->count, sizeof(char *));
-	content->count = splitLines(content->buffer, content->lines, content->count);
-
-	if (content->lines == NULL)
+	if (!splitLines(&(content->lbuf), contents))
 	{
-		log_error(ALLOCATION_FAILED_ERROR);
+		/* errors have already been logged */
 		return false;
 	}
 
+	/*
+	 * If the file contains zero lines, we're done already, Also malloc(zero)
+	 * leads to "corrupted size vs. prev_size" run-time errors.
+	 */
+	if (content->lbuf.count == 0)
+	{
+		return true;
+	}
+
 	content->messages =
-		(LogicalMessageMetadata *) calloc(content->count,
+		(LogicalMessageMetadata *) calloc(content->lbuf.count,
 										  sizeof(LogicalMessageMetadata));
 
 	if (content->messages == NULL)
@@ -1958,9 +1989,9 @@ stream_read_file(StreamContent *content)
 		return false;
 	}
 
-	for (int i = 0; i < content->count; i++)
+	for (uint64_t i = 0; i < content->lbuf.count; i++)
 	{
-		char *message = content->lines[i];
+		char *message = content->lbuf.lines[i];
 		LogicalMessageMetadata *metadata = &(content->messages[i]);
 
 		JSON_Value *json = json_parse_string(message);
@@ -1968,11 +1999,8 @@ stream_read_file(StreamContent *content)
 		if (!parseMessageMetadata(metadata, message, json, false))
 		{
 			/* errors have already been logged */
-			json_value_free(json);
 			return false;
 		}
-
-		json_value_free(json);
 	}
 
 	return true;
@@ -2135,11 +2163,9 @@ buildReplicationURI(const char *pguri, char **repl_pguri)
 	if (!buildPostgresURIfromPieces(&params, repl_pguri))
 	{
 		log_error("Failed to produce the replication connection string");
-		freeURIParams(&params);
 		return false;
 	}
 
-	freeURIParams(&params);
 	return true;
 }
 
@@ -2752,33 +2778,21 @@ stream_read_context(CDCPaths *paths,
 	if (!stringToUInt(wal_segment_size, WalSegSz))
 	{
 		/* errors have already been logged */
-		free(wal_segment_size);
-		free(tli);
-		free(history);
 		return false;
 	}
 
 	if (!stringToUInt(tli, &(system->timeline)))
 	{
 		/* errors have already been logged */
-		free(wal_segment_size);
-		free(tli);
-		free(history);
 		return false;
 	}
 
 	if (!parseTimeLineHistory(paths->tlihistfile, history, system))
 	{
 		/* errors have already been logged */
-		free(wal_segment_size);
-		free(tli);
-		free(history);
 		return false;
 	}
 
-	free(wal_segment_size);
-	free(tli);
-	free(history);
 
 	return true;
 }
