@@ -14,6 +14,7 @@
 #include "queue_utils.h"
 #include "pgsql.h"
 #include "schema.h"
+#include "string_utils.h"
 
 #define OUTPUT_BEGIN "BEGIN; -- "
 #define OUTPUT_COMMIT "COMMIT; -- "
@@ -139,7 +140,6 @@ typedef struct LogicalMessageRelation
 {
 	char *nspname;  /* malloc'ed area */
 	char *relname;  /* malloc'ed area */
-	bool pqMemory;
 } LogicalMessageRelation;
 
 typedef struct LogicalMessageInsert
@@ -361,8 +361,11 @@ typedef struct StreamApplyContext
 {
 	CDCPaths paths;
 
-	/* target connection */
-	PGSQL pgsql;
+	/* target connection to find current wal_lsn for replay_lsn mapping */
+	PGSQL controlPgConn;
+
+	/* target connection created in pipeline mode responsible for apply */
+	PGSQL applyPgConn;
 
 	/* apply needs access to the catalogs to register sentinel replay_lsn */
 	DatabaseCatalog *sourceDB;
@@ -401,9 +404,7 @@ typedef struct StreamApplyContext
 typedef struct StreamContent
 {
 	char filename[MAXPGPATH];
-	int count;
-	char *buffer;
-	char **lines;                     /* malloc'ed area */
+	LinesBuffer lbuf;
 	LogicalMessageMetadata *messages; /* malloc'ed area */
 } StreamContent;
 
@@ -625,11 +626,6 @@ bool parseMessage(StreamContext *privateContext, char *message, JSON_Value *json
 bool streamLogicalTransactionAppendStatement(LogicalTransaction *txn,
 											 LogicalTransactionStatement *stmt);
 
-void FreeLogicalMessage(LogicalMessage *msg);
-void FreeLogicalTransaction(LogicalTransaction *tx);
-void FreeLogicalMessageTupleArray(LogicalMessageTupleArray *tupleArray);
-void FreeLogicalMessageRelation(LogicalMessageRelation *table);
-void FreeLogicalMessageTuple(LogicalMessageTuple *tuple);
 bool AllocateLogicalMessageTuple(LogicalMessageTuple *tuple, int count);
 
 /* ld_test_decoding.c */
@@ -655,6 +651,8 @@ bool stream_apply_catchup(StreamSpecs *specs);
 
 bool stream_apply_setup(StreamSpecs *specs, StreamApplyContext *context);
 
+bool stream_apply_cleanup(StreamApplyContext *context);
+
 bool stream_apply_wait_for_sentinel(StreamSpecs *specs,
 									StreamApplyContext *context);
 
@@ -680,14 +678,9 @@ bool computeSQLFileName(StreamApplyContext *context);
 
 bool parseSQLAction(const char *query, LogicalMessageMetadata *metadata);
 
-bool stream_apply_track_insert_lsn(StreamApplyContext *context,
-								   uint64_t sourceLSN);
-
 bool stream_apply_find_durable_lsn(StreamApplyContext *context,
 								   uint64_t *durableLSN);
 
-bool stream_apply_write_lsn_tracking(StreamApplyContext *context);
-bool stream_apply_read_lsn_tracking(StreamApplyContext *context);
 
 /* ld_replay */
 bool stream_replay(StreamSpecs *specs);
