@@ -110,7 +110,6 @@ prepareTestDecodingMessage(LogicalStreamContext *context)
 		return false;
 	}
 
-	json_value_free(js);
 	json_free_serialized_string(jsonstr);
 
 	return true;
@@ -352,7 +351,6 @@ parseTestDecodingMessageHeader(TestDecodingHeader *header, const char *message)
 	 */
 	header->table.nspname = strndup(idp, dot - idp);
 	header->table.relname = strndup(dot + 1, sep - dot - 1);
-	header->table.pqMemory = false;
 
 	sformat(header->qname, sizeof(header->qname), "%s.%s",
 			header->table.nspname,
@@ -627,21 +625,6 @@ SetColumnNamesAndValues(LogicalMessageTuple *tuple, TestDecodingHeader *header)
 		return false;
 	}
 
-	/*
-	 * Free the TestDecodingColumns memory that we allocated: only the
-	 * structure itself, the rest is just a bunch of pointers to parts of the
-	 * messages.
-	 */
-	TestDecodingColumns *c = cols;
-
-	for (; c != NULL;)
-	{
-		TestDecodingColumns *next = c->next;
-
-		free(c);
-		c = next;
-	}
-
 	return true;
 }
 
@@ -719,7 +702,7 @@ parseNextColumn(TestDecodingColumns *cols,
 	cols->colnameStart = ptr;
 	cols->colnameLen = typA - ptr;
 
-	log_trace("parseNextColumn[%s]: %.*s",
+	log_trace("parseNextColumn[%s]: \"%.*s\"",
 			  typname,
 			  cols->colnameLen,
 			  cols->colnameStart);
@@ -798,7 +781,7 @@ parseNextColumn(TestDecodingColumns *cols,
 
 		/* advance to past the value, skip the next space */
 		ptr = end + 1;
-		header->pos = ptr - header->message;
+		header->pos = ptr - header->message + 1;
 
 		log_trace("parseNextColumn: bit string value: %.*s",
 				  cols->valueLen,
@@ -984,7 +967,6 @@ prepareUpdateTuppleArrays(StreamContext *privateContext,
 										table))
 	{
 		/* errors have already been logged */
-		free(table);
 		return false;
 	}
 
@@ -1006,41 +988,52 @@ prepareUpdateTuppleArrays(StreamContext *privateContext,
 	}
 
 	int columnCount = cols->values.array[0].cols;
-	bool *pkeyArray = (bool *) calloc(columnCount, sizeof(bool));
+	bool *pkeyArray = NULL;
 
 	int oldCount = 0;
 	int newCount = 0;
 
-	for (int c = 0; c < columnCount; c++)
+	if (0 < columnCount)
 	{
-		const char *colname = cols->columns[c];
+		pkeyArray = (bool *) calloc(columnCount, sizeof(bool));
 
-		SourceTableAttribute attribute = { 0 };
-
-		if (!catalog_lookup_s_attr_by_name(sourceDB,
-										   table->oid,
-										   colname,
-										   &attribute))
+		if (pkeyArray == NULL)
 		{
-			log_error("Failed to lookup for table %s attribute %s in our "
-					  "internal catalogs, see above for details",
-					  table->qname,
-					  colname);
+			log_error(ALLOCATION_FAILED_ERROR);
 			return false;
 		}
 
-		if (attribute.attnum > 0)
+		for (int c = 0; c < columnCount; c++)
 		{
-			pkeyArray[c] = attribute.attisprimary;
-		}
+			const char *colname = cols->columns[c];
 
-		if (pkeyArray[c])
-		{
-			++oldCount;
-		}
-		else
-		{
-			++newCount;
+			SourceTableAttribute attribute = { 0 };
+
+			if (!catalog_lookup_s_attr_by_name(sourceDB,
+											   table->oid,
+											   colname,
+											   &attribute))
+			{
+				log_error("Failed to lookup for table %s attribute %s in our "
+						  "internal catalogs, see above for details",
+						  table->qname,
+						  colname);
+				return false;
+			}
+
+			if (attribute.attnum > 0)
+			{
+				pkeyArray[c] = attribute.attisprimary;
+			}
+
+			if (pkeyArray[c])
+			{
+				++oldCount;
+			}
+			else
+			{
+				++newCount;
+			}
 		}
 	}
 
@@ -1110,7 +1103,5 @@ prepareUpdateTuppleArrays(StreamContext *privateContext,
 		cols->values.array[0].array[c].val.str = NULL;
 	}
 
-	(void) FreeLogicalMessageTuple(cols);
-
 	return true;
 }
