@@ -456,8 +456,6 @@ stream_transform_write_message(StreamContext *privateContext,
 		return false;
 	}
 
-	(void) FreeLogicalMessage(currentMsg);
-
 	if (metadata->action == STREAM_ACTION_COMMIT ||
 		metadata->action == STREAM_ACTION_ROLLBACK)
 	{
@@ -515,7 +513,6 @@ stream_transform_message(StreamContext *privateContext, char *message)
 	if (!parseMessageMetadata(metadata, message, json, false))
 	{
 		/* errors have already been logged */
-		json_value_free(json);
 		return false;
 	}
 
@@ -524,11 +521,9 @@ stream_transform_message(StreamContext *privateContext, char *message)
 		log_error("Failed to parse JSON message: %.1024s%s",
 				  message,
 				  strlen(message) > 1024 ? "..." : "");
-		json_value_free(json);
 		return false;
 	}
 
-	json_value_free(json);
 
 	return true;
 }
@@ -886,7 +881,6 @@ stream_transform_file(StreamSpecs *specs, char *jsonfilename, char *sqlfilename)
 {
 	StreamContext *privateContext = &(specs->private);
 	StreamContent content = { 0 };
-	long size = 0L;
 
 	log_notice("Transforming JSON file \"%s\" into SQL file \"%s\"",
 			   jsonfilename,
@@ -899,28 +893,34 @@ stream_transform_file(StreamSpecs *specs, char *jsonfilename, char *sqlfilename)
 	 * decoding messages, and parse the JSON messages into our internal
 	 * representation structure.
 	 */
-	if (!read_file(content.filename, &(content.buffer), &size))
+	char *contents = NULL;
+	long size = 0L;
+
+	if (!read_file(content.filename, &contents, &size))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
-	content.count = countLines(content.buffer);
-	content.lines = (char **) calloc(content.count, sizeof(char *));
-	content.count = splitLines(content.buffer, content.lines, content.count);
-
-	if (content.lines == NULL)
+	if (!splitLines(&(content.lbuf), contents))
 	{
-		log_error(ALLOCATION_FAILED_ERROR);
-		free(content.buffer);
-		free(content.lines);
+		/* errors have already been logged */
 		return false;
 	}
 
-	log_debug("stream_transform_file: read %d lines from \"%s\"",
-			  content.count,
+	log_debug("stream_transform_file: read %lld lines from \"%s\"",
+			  (long long) content.lbuf.count,
 			  content.filename);
 
+	/*
+	 * If the file contains zero lines, we're done already, Also malloc(zero)
+	 * leads to "corrupted size vs. prev_size" run-time errors.
+	 */
+	if (content.lbuf.count == 0)
+	{
+		return true;
+	}
+
 	/*
 	 * The output is written to a temp/partial file which is renamed after
 	 * close, so that another tool that would want to read the file won't read
@@ -936,8 +936,6 @@ stream_transform_file(StreamSpecs *specs, char *jsonfilename, char *sqlfilename)
 	if (privateContext->sqlFile == NULL)
 	{
 		log_error("Failed to open file \"%s\"", tempfilename);
-		free(content.buffer);
-		free(content.lines);
 		return false;
 	}
 
@@ -952,23 +950,20 @@ stream_transform_file(StreamSpecs *specs, char *jsonfilename, char *sqlfilename)
 	/* we skip KEEPALIVE message in the beginning of the file */
 	bool firstMessage = true;
 
-	for (int i = 0; i < content.count; i++)
+	for (uint64_t i = 0; i < content.lbuf.count; i++)
 	{
-		char *message = content.lines[i];
+		char *message = content.lbuf.lines[i];
 
 		LogicalMessageMetadata empty = { 0 };
 		*metadata = empty;
 
-		log_trace("stream_transform_file[%2d]: %s", i, message);
+		log_trace("stream_transform_file[%4lld]: %s", (long long) i, message);
 
 		JSON_Value *json = json_parse_string(message);
 
 		if (!parseMessageMetadata(metadata, message, json, false))
 		{
 			/* errors have already been logged */
-			json_value_free(json);
-			free(content.buffer);
-			free(content.lines);
 			return false;
 		}
 
@@ -991,6 +986,12 @@ stream_transform_file(StreamSpecs *specs, char *jsonfilename, char *sqlfilename)
 
 			LogicalTransaction *txn = &(new.command.tx);
 			txn->continued = true;
+
+			/*
+			 * test_decoding DML logical messages will always have xid = 0.
+			 * We handle that in parseMessage STREAM_ACTION_COMMIT by using
+			 * the xid from the COMMIT message.
+			 */
 			txn->xid = metadata->xid;
 			txn->first = NULL;
 
@@ -1000,13 +1001,9 @@ stream_transform_file(StreamSpecs *specs, char *jsonfilename, char *sqlfilename)
 		if (!parseMessage(privateContext, message, json))
 		{
 			log_error("Failed to parse JSON message: %s", message);
-			json_value_free(json);
-			free(content.buffer);
-			free(content.lines);
 			return false;
 		}
 
-		json_value_free(json);
 
 		/*
 		 * Prepare a new message when we just read the COMMIT message of an
@@ -1018,8 +1015,6 @@ stream_transform_file(StreamSpecs *specs, char *jsonfilename, char *sqlfilename)
 		{
 			log_error("Failed to transform and flush the current message, "
 					  "see above for details");
-			free(content.buffer);
-			free(content.lines);
 			return false;
 		}
 
@@ -1035,9 +1030,7 @@ stream_transform_file(StreamSpecs *specs, char *jsonfilename, char *sqlfilename)
 
 	if (fclose(privateContext->sqlFile) == EOF)
 	{
-		log_error("Failed to write file \"%s\"", tempfilename);
-		free(content.buffer);
-		free(content.lines);
+		log_error("Failed to close file \"%s\"", tempfilename);
 		return false;
 	}
 
@@ -1052,18 +1045,13 @@ stream_transform_file(StreamSpecs *specs, char *jsonfilename, char *sqlfilename)
 		log_error("Failed to move \"%s\" to \"%s\": %m",
 				  tempfilename,
 				  sqlfilename);
-		free(content.buffer);
-		free(content.lines);
 		return false;
 	}
 
-	log_info("Transformed %d JSON messages into SQL file \"%s\"",
-			 content.count,
+	log_info("Transformed %lld JSON messages into SQL file \"%s\"",
+			 (long long) content.lbuf.count,
 			 sqlfilename);
 
-	free(content.buffer);
-	free(content.lines);
-
 	return true;
 }
 
@@ -1216,6 +1204,13 @@ parseMessage(StreamContext *privateContext, char *message, JSON_Value *json)
 			/* update the timestamp for tracking in replication origin */
 			strlcpy(txn->timestamp, metadata->timestamp, sizeof(txn->timestamp));
 			txn->commitLSN = metadata->lsn;
+
+			/*
+			 * Unlike wal2json, test_decoding don't have xid in the DML logical
+			 * messages. So we use the xid from the COMMIT message to update the
+			 * transaction xid.
+			 */
+			txn->xid = metadata->xid;
 			txn->commit = true;
 
 			break;
@@ -1248,7 +1243,6 @@ parseMessage(StreamContext *privateContext, char *message, JSON_Value *json)
 				/* copy the stmt over, then free the extra allocated memory */
 				mesg->action = metadata->action;
 				mesg->command.switchwal = stmt->stmt.switchwal;
-				free(stmt);
 			}
 
 			break;
@@ -1272,7 +1266,6 @@ parseMessage(StreamContext *privateContext, char *message, JSON_Value *json)
 				/* copy the stmt over, then free the extra allocated memory */
 				mesg->action = metadata->action;
 				mesg->command.keepalive = stmt->stmt.keepalive;
-				free(stmt);
 			}
 
 			break;
@@ -1291,7 +1284,6 @@ parseMessage(StreamContext *privateContext, char *message, JSON_Value *json)
 				/* copy the stmt over, then free the extra allocated memory */
 				mesg->action = metadata->action;
 				mesg->command.endpos = stmt->stmt.endpos;
-				free(stmt);
 			}
 
 			break;
@@ -1425,10 +1417,6 @@ coalesceLogicalTransactionStatement(LogicalTransaction *txn,
 	lastValuesArray->array[lastValuesArray->count++] = newValuesArray->array[0];
 	newValuesArray->count = 0;
 
-	/* Deallocate the tuple and the new statement */
-	FreeLogicalMessageTupleArray(&(new->stmt.insert.new));
-	free(new);
-
 	return true;
 }
 
@@ -1569,151 +1557,6 @@ streamLogicalTransactionAppendStatement(LogicalTransaction *txn,
 }
 
 
-/*
- * FreeLogicalMessage frees the malloc'ated memory areas of a LogicalMessage.
- */
-void
-FreeLogicalMessage(LogicalMessage *msg)
-{
-	if (msg->isTransaction)
-	{
-		FreeLogicalTransaction(&(msg->command.tx));
-	}
-}
-
-
-/*
- * FreeLogicalTransaction frees the malloc'ated memory areas of a
- * LogicalTransaction.
- */
-void
-FreeLogicalTransaction(LogicalTransaction *tx)
-{
-	LogicalTransactionStatement *currentStmt = tx->first;
-
-	for (; currentStmt != NULL;)
-	{
-		switch (currentStmt->action)
-		{
-			case STREAM_ACTION_INSERT:
-			{
-				FreeLogicalMessageRelation(&(currentStmt->stmt.insert.table));
-				FreeLogicalMessageTupleArray(&(currentStmt->stmt.insert.new));
-				break;
-			}
-
-			case STREAM_ACTION_UPDATE:
-			{
-				FreeLogicalMessageRelation(&(currentStmt->stmt.update.table));
-				FreeLogicalMessageTupleArray(&(currentStmt->stmt.update.old));
-				FreeLogicalMessageTupleArray(&(currentStmt->stmt.update.new));
-				break;
-			}
-
-			case STREAM_ACTION_DELETE:
-			{
-				FreeLogicalMessageRelation(&(currentStmt->stmt.delete.table));
-				FreeLogicalMessageTupleArray(&(currentStmt->stmt.delete.old));
-				break;
-			}
-
-			case STREAM_ACTION_TRUNCATE:
-			{
-				FreeLogicalMessageRelation(&(currentStmt->stmt.truncate.table));
-				break;
-			}
-
-			/* no malloc'ated area in a BEGIN, COMMIT, or TRUNCATE statement */
-			default:
-			{
-				break;
-			}
-		}
-
-		LogicalTransactionStatement *stmt = currentStmt;
-		currentStmt = currentStmt->next;
-
-		free(stmt);
-	}
-
-	tx->first = NULL;
-}
-
-
-/*
- * FreeLogicalMessageRelation frees the malloc'ated memory areas of
- * LogicalMessageRelation.
- */
-void
-FreeLogicalMessageRelation(LogicalMessageRelation *table)
-{
-	if (table->pqMemory)
-	{
-		/* use PQfreemem for memory allocated by PQescapeIdentifer */
-		PQfreemem(table->nspname);
-		PQfreemem(table->relname);
-	}
-	else
-	{
-		free(table->nspname);
-		free(table->relname);
-	}
-}
-
-
-/*
- * FreeLogicalMessageTupleArray frees the malloc'ated memory areas of a
- * LogicalMessageTupleArray.
- */
-void
-FreeLogicalMessageTupleArray(LogicalMessageTupleArray *tupleArray)
-{
-	for (int s = 0; s < tupleArray->count; s++)
-	{
-		LogicalMessageTuple *tuple = &(tupleArray->array[s]);
-
-		(void) FreeLogicalMessageTuple(tuple);
-	}
-
-	free(tupleArray->array);
-}
-
-
-/*
- * FreeLogicalMessageTuple frees the malloc'ated memory areas of a
- * LogicalMessageTuple.
- */
-void
-FreeLogicalMessageTuple(LogicalMessageTuple *tuple)
-{
-	for (int i = 0; i < tuple->cols; i++)
-	{
-		free(tuple->columns[i]);
-	}
-	free(tuple->columns);
-
-	for (int r = 0; r < tuple->values.count; r++)
-	{
-		LogicalMessageValues *values = &(tuple->values.array[r]);
-
-		for (int v = 0; v < values->cols; v++)
-		{
-			LogicalMessageValue *value = &(values->array[v]);
-
-			if ((value->oid == TEXTOID || value->oid == BYTEAOID) &&
-				!value->isNull)
-			{
-				free(value->val.str);
-			}
-		}
-
-		free(values->array);
-	}
-
-	free(tuple->values.array);
-}
-
-
 /*
  * allocateLogicalMessageTuple allocates memory for count columns (and values)
  * for the given LogicalMessageTuple.
@@ -1722,6 +1565,19 @@ bool
 AllocateLogicalMessageTuple(LogicalMessageTuple *tuple, int count)
 {
 	tuple->cols = count;
+
+	if (count == 0)
+	{
+		tuple->columns = NULL;
+
+		LogicalMessageValuesArray *valuesArray = &(tuple->values);
+		valuesArray->count = 0;
+		valuesArray->capacity = 0;
+		valuesArray->array = NULL;
+
+		return true;
+	}
+
 	tuple->columns = (char **) calloc(count, sizeof(char *));
 
 	if (tuple->columns == NULL)
@@ -2178,14 +2034,7 @@ stream_write_insert(FILE *out, LogicalMessageInsert *insert)
 
 		for (int c = 0; c < stmt->cols; c++)
 		{
-			/*
-			 * In the case of the test_decoding plugin, it already escapes
-			 * keywords using double quotes, so we should avoid double quoting
-			 * again.
-			 */
-			const char *quoteFormatStr = (stmt->columns[c][0] == '"') ? "%s%s" :
-										 "%s\"%s\"";
-			appendPQExpBuffer(buf, quoteFormatStr,
+			appendPQExpBuffer(buf, "%s%s",
 							  c > 0 ? ", " : "",
 							  stmt->columns[c]);
 		}
@@ -2258,7 +2107,6 @@ stream_write_insert(FILE *out, LogicalMessageInsert *insert)
 		FFORMAT(out, "EXECUTE %x%s;\n", hash, serialized_string);
 
 		json_free_serialized_string(serialized_string);
-		json_value_free(js);
 	}
 
 	return true;
@@ -2358,14 +2206,7 @@ stream_write_update(FILE *out, LogicalMessageUpdate *update)
 
 				if (!skip)
 				{
-					/*
-					 * In the case of the test_decoding plugin, it already escapes
-					 * keywords using double quotes, so we should avoid double quoting
-					 * again.
-					 */
-					const char *quoteFormatStr = (colname[0] == '"') ? "%s%s = $%d" :
-												 "%s\"%s\" = $%d";
-					appendPQExpBuffer(buf, quoteFormatStr,
+					appendPQExpBuffer(buf, "%s%s = $%d",
 									  first ? "" : ", ",
 									  colname,
 									  ++pos);
@@ -2406,23 +2247,29 @@ stream_write_update(FILE *out, LogicalMessageUpdate *update)
 					return false;
 				}
 
-				/*
-				 * In the case of the test_decoding plugin, it already escapes
-				 * keywords using double quotes, so we should avoid double quoting
-				 * again.
-				 */
-				const char *quoteFormatStr = (old->columns[v][0] == '"') ? "%s%s = $%d" :
-											 "%s\"%s\" = $%d";
-				appendPQExpBuffer(buf, quoteFormatStr,
-								  v > 0 ? " and " : "",
-								  old->columns[v],
-								  ++pos);
-
-				if (!stream_add_value_in_json_array(value, jsArray))
+				if (value->isNull)
 				{
-					/* errors have already been logged */
-					destroyPQExpBuffer(buf);
-					return false;
+					/*
+					 * Attributes with the value `NULL` require `IS NULL` instead of `=`
+					 * in the WHERE clause.
+					 */
+					appendPQExpBuffer(buf, "%s%s IS NULL",
+									  v > 0 ? " and " : "",
+									  old->columns[v]);
+				}
+				else
+				{
+					appendPQExpBuffer(buf, "%s%s = $%d",
+									  v > 0 ? " and " : "",
+									  old->columns[v],
+									  ++pos);
+
+					if (!stream_add_value_in_json_array(value, jsArray))
+					{
+						/* errors have already been logged */
+						destroyPQExpBuffer(buf);
+						return false;
+					}
 				}
 			}
 		}
@@ -2448,7 +2295,6 @@ stream_write_update(FILE *out, LogicalMessageUpdate *update)
 		FFORMAT(out, "EXECUTE %x%s;\n", hash, serialized_string);
 
 		json_free_serialized_string(serialized_string);
-		json_value_free(js);
 	}
 
 	return true;
@@ -2499,23 +2345,29 @@ stream_write_delete(FILE *out, LogicalMessageDelete *delete)
 					return false;
 				}
 
-				/*
-				 * In the case of the test_decoding plugin, it already escapes
-				 * keywords using double quotes, so we should avoid double quoting
-				 * again.
-				 */
-				const char *quoteFormatStr = (old->columns[v][0] == '"') ? "%s%s = $%d" :
-											 "%s\"%s\" = $%d";
-				appendPQExpBuffer(buf, quoteFormatStr,
-								  v > 0 ? " and " : "",
-								  old->columns[v],
-								  ++pos);
-
-				if (!stream_add_value_in_json_array(value, jsArray))
+				if (value->isNull)
 				{
-					/* errors have already been logged */
-					destroyPQExpBuffer(buf);
-					return false;
+					/*
+					 * Attributes with the value `NULL` require `IS NULL` instead of `=`
+					 * in the WHERE clause.
+					 */
+					appendPQExpBuffer(buf, "%s%s IS NULL",
+									  v > 0 ? " and " : "",
+									  old->columns[v]);
+				}
+				else
+				{
+					appendPQExpBuffer(buf, "%s%s = $%d",
+									  v > 0 ? " and " : "",
+									  old->columns[v],
+									  ++pos);
+
+					if (!stream_add_value_in_json_array(value, jsArray))
+					{
+						/* errors have already been logged */
+						destroyPQExpBuffer(buf);
+						return false;
+					}
 				}
 			}
 		}
@@ -2534,7 +2386,6 @@ stream_write_delete(FILE *out, LogicalMessageDelete *delete)
 		FFORMAT(out, "EXECUTE %x%s;\n", hash, serialized_string);
 
 		json_free_serialized_string(serialized_string);
-		json_value_free(js);
 	}
 
 	return true;
