@@ -28,11 +28,17 @@
 #include "signals.h"
 #include "string_utils.h"
 
+/*
+ * Because we did include defaults.h previously in the same compilation unit (a
+ * .c file), then the runprogram.h code is linked to the Garbage-Collector:
+ * calls to malloc() in there are replaced by calls to GC_malloc() as per
+ * defaults.h #define instructions.
+ *
+ * As a result we never call free_program().
+ */
 #define RUN_PROGRAM_IMPLEMENTATION
 #include "runprogram.h"
 
-static bool pg_dump_db_extension_namespace_hook(void *ctx, SourceExtension *ext);
-
 
 /*
  * Get psql --version output in pgPaths->pg_version.
@@ -40,29 +46,32 @@ static bool pg_dump_db_extension_namespace_hook(void *ctx, SourceExtension *ext)
 bool
 psql_version(PostgresPaths *pgPaths)
 {
-	Program prog = run_program(pgPaths->psql, "--version", NULL);
+	Program *prog = run_program(pgPaths->psql, "--version", NULL);
 	char pg_version_string[PG_VERSION_STRING_MAX] = { 0 };
 	int pg_version = 0;
 
-	if (prog.returnCode != 0)
+	if (prog == NULL)
 	{
-		errno = prog.error;
-		log_error("Failed to run \"psql --version\" using program \"%s\": %m",
-				  pgPaths->psql);
-		free_program(&prog);
+		log_error(ALLOCATION_FAILED_ERROR);
 		return false;
 	}
 
-	if (!parse_version_number(prog.stdOut,
+	if (prog->returnCode != 0)
+	{
+		errno = prog->error;
+		log_error("Failed to run \"psql --version\" using program \"%s\": %m",
+				  pgPaths->psql);
+		return false;
+	}
+
+	if (!parse_version_number(prog->stdOut,
 							  pg_version_string,
 							  PG_VERSION_STRING_MAX,
 							  &pg_version))
 	{
 		/* errors have already been logged */
-		free_program(&prog);
 		return false;
 	}
-	free_program(&prog);
 
 	strlcpy(pgPaths->pg_version, pg_version_string, PG_VERSION_STRING_MAX);
 
@@ -193,32 +202,33 @@ set_psql_from_config_bindir(PostgresPaths *pgPaths, const char *pg_config)
 		return false;
 	}
 
-	Program prog = run_program(pg_config, "--bindir", NULL);
+	Program *prog = run_program(pg_config, "--bindir", NULL);
 
-	char *lines[1];
-
-	if (prog.returnCode != 0)
+	if (prog == NULL)
 	{
-		errno = prog.error;
+		log_error(ALLOCATION_FAILED_ERROR);
+		return false;
+	}
+
+	if (prog->returnCode != 0)
+	{
+		errno = prog->error;
 		log_error("Failed to run \"pg_config --bindir\" using program \"%s\": %m",
 				  pg_config);
-		free_program(&prog);
 		return false;
 	}
 
-	if (splitLines(prog.stdOut, lines, 1) != 1)
+	LinesBuffer lbuf = { 0 };
+
+	if (!splitLines(&lbuf, prog->stdOut) || lbuf.count != 1)
 	{
 		log_error("Unable to parse output from pg_config --bindir");
-		free_program(&prog);
 		return false;
 	}
 
-	char *bindir = lines[0];
+	char *bindir = lbuf.lines[0];
 	join_path_components(psql, bindir, "psql");
 
-	/* we're now done with the Program and its output */
-	free_program(&prog);
-
 	if (!file_exists(psql))
 	{
 		log_error("Failed to find psql at \"%s\" from PG_CONFIG at \"%s\"",
@@ -431,46 +441,6 @@ pg_dump_db(PostgresPaths *pgPaths,
 		args[argsIndex++] = nspname;
 	}
 
-	/*
-	 * Store extension args in a separate array, extension args will dynamically
-	 * allocated by pg_dump_db_extension_namespace_hook and we want to free
-	 * them after the call to pg_dump.
-	 */
-	char *extNamespaces[PG_CMD_MAX_ARG];
-	int extNamespaceCount = 0;
-
-	/* now --exclude-schema for extension's own schemas */
-	DumpExtensionNamespaceContext context = {
-		.extNamespaces = extNamespaces,
-		.extNamespaceCount = &extNamespaceCount,
-	};
-
-	if (!catalog_iter_s_extension(filtersDB,
-								  &context,
-								  &pg_dump_db_extension_namespace_hook))
-	{
-		log_error("Failed to prepare pg_dump command line arguments, "
-				  "see above for details");
-		return false;
-	}
-
-	for (int i = 0; i < extNamespaceCount; i++)
-	{
-		/* check that we still have room for --exclude-schema args */
-		if (PG_CMD_MAX_ARG < (argsIndex + 2))
-		{
-			log_error("Failed to call pg_dump, "
-					  "too many schema are excluded: "
-					  "argsIndex %d > %d",
-					  argsIndex + 2,
-					  PG_CMD_MAX_ARG);
-			return false;
-		}
-
-		args[argsIndex++] = "--exclude-schema";
-		args[argsIndex++] = extNamespaces[i];
-	}
-
 	args[argsIndex++] = "--file";
 	args[argsIndex++] = (char *) filename;
 	args[argsIndex++] = (char *) connStrings->safeSourcePGURI.pguri;
@@ -485,12 +455,6 @@ pg_dump_db(PostgresPaths *pgPaths,
 	(void) initialize_program(&program, args, false);
 	program.processBuffer = &processBufferCallback;
 
-	/* free the extension namespace string values */
-	for (int i = 0; i < extNamespaceCount; i++)
-	{
-		free(extNamespaces[i]);
-	}
-
 	/* log the exact command line we're using */
 	int commandSize = snprintf_program_command_line(&program, command, BUFSIZE);
 
@@ -516,45 +480,10 @@ pg_dump_db(PostgresPaths *pgPaths,
 	if (program.returnCode != 0)
 	{
 		log_error("Failed to run pg_dump: exit code %d", program.returnCode);
-		free_program(&program);
 
 		return false;
 	}
 
-	free_program(&program);
-	return true;
-}
-
-
-/*
- * pg_dump_db_extension_namespace_hook is an iterator callback function.
- */
-static bool
-pg_dump_db_extension_namespace_hook(void *ctx, SourceExtension *ext)
-{
-	DumpExtensionNamespaceContext *context =
-		(DumpExtensionNamespaceContext *) ctx;
-
-	char **extNamespaces = context->extNamespaces;
-
-	char *nspname = ext->extnamespace;
-
-	if (!streq(nspname, "public") && !streq(nspname, "pg_catalog"))
-	{
-		/* check that we still have room for args */
-		if (PG_CMD_MAX_ARG < ((*context->extNamespaceCount) + 1))
-		{
-			log_error("Failed to call pg_dump, "
-					  "too many schema are excluded: "
-					  "extNamespaceCount %d > %d",
-					  (*context->extNamespaceCount) + 1,
-					  PG_CMD_MAX_ARG);
-			return false;
-		}
-
-		extNamespaces[(*context->extNamespaceCount)++] = strdup(nspname);
-	}
-
 	return true;
 }
 
@@ -641,12 +570,10 @@ pg_dumpall_roles(PostgresPaths *pgPaths,
 	if (program.returnCode != 0)
 	{
 		log_error("Failed to run pg_dump: exit code %d", program.returnCode);
-		free_program(&program);
 
 		return false;
 	}
 
-	free_program(&program);
 	return true;
 }
 
@@ -691,8 +618,13 @@ pg_restore_roles(PostgresPaths *pgPaths,
 		return false;
 	}
 
-	char *lines[BUFSIZE] = { 0 };
-	int lineCount = splitLines(content, lines, BUFSIZE);
+	LinesBuffer lbuf = { 0 };
+
+	if (!splitLines(&lbuf, content))
+	{
+		/* errors have already been logged */
+		return false;
+	}
 
 	PGSQL pgsql = { 0 };
 
@@ -718,9 +650,9 @@ pg_restore_roles(PostgresPaths *pgPaths,
 	 */
 	bool skipNextLine = false;
 
-	for (int l = 0; l < lineCount; l++)
+	for (uint64_t l = 0; l < lbuf.count; l++)
 	{
-		char *currentLine = lines[l];
+		char *currentLine = lbuf.lines[l];
 
 		if (skipNextLine)
 		{
@@ -915,6 +847,11 @@ pg_restore_db(PostgresPaths *pgPaths,
 		args[argsIndex++] = "--no-acl";
 	}
 
+	if (options.noTableSpaces)
+	{
+		args[argsIndex++] = "--no-tablespaces";
+	}
+
 	/*
 	 * Do not apply [include-only-schema] filtering.
 	 *
@@ -986,12 +923,10 @@ pg_restore_db(PostgresPaths *pgPaths,
 	if (program.returnCode != 0)
 	{
 		log_error("Failed to run pg_restore: exit code %d", program.returnCode);
-		free_program(&program);
 
 		return false;
 	}
 
-	free_program(&program);
 	return true;
 }
 
@@ -1038,7 +973,6 @@ pg_restore_list(PostgresPaths *pgPaths,
 	if (program.returnCode != 0)
 	{
 		log_error("Failed to run pg_restore: exit code %d", program.returnCode);
-		free_program(&program);
 
 		return false;
 	}
@@ -1046,11 +980,9 @@ pg_restore_list(PostgresPaths *pgPaths,
 	if (!parse_archive_list(listFilename, archive))
 	{
 		/* errors have already been logged */
-		free_program(&program);
 		return false;
 	}
 
-	free_program(&program);
 	return true;
 }
 
@@ -1096,6 +1028,7 @@ struct ArchiveItemDescMapping pgRestoreDescriptionArray[] = {
 	INSERT_MAPPING(ARCHIVE_TAG_COMMENT, "COMMENT"),
 	INSERT_MAPPING(ARCHIVE_TAG_CONSTRAINT, "CONSTRAINT"),
 	INSERT_MAPPING(ARCHIVE_TAG_CONVERSION, "CONVERSION"),
+	INSERT_MAPPING(ARCHIVE_TAG_DATABASE, "DATABASE"),
 	INSERT_MAPPING(ARCHIVE_TAG_DEFAULT_ACL, "DEFAULT ACL"),
 	INSERT_MAPPING(ARCHIVE_TAG_DEFAULT, "DEFAULT"),
 	INSERT_MAPPING(ARCHIVE_TAG_DOMAIN, "DOMAIN"),
@@ -1164,28 +1097,32 @@ parse_archive_list(const char *filename, ArchiveContentArray *contents)
 		return false;
 	}
 
-	int lineCount = countLines(buffer);
-	char **lines = (char **) calloc(lineCount, sizeof(char *));
-	int splitCount = splitLines(buffer, lines, lineCount);
+	LinesBuffer lbuf = { 0 };
 
-	if (splitCount != lineCount)
+	if (!splitLines(&lbuf, buffer))
 	{
-		log_error("BUG: parse_archive_list counted %d lines "
-				  "and got %d after split",
-				  lineCount,
-				  splitCount);
+		/* errors have already been logged */
 		return false;
 	}
 
+	/*
+	 * If the file contains zero lines, we're done already, Also malloc(zero)
+	 * leads to "corrupted size vs. prev_size" run-time errors.
+	 */
+	if (lbuf.count == 0)
+	{
+		return true;
+	}
+
 	contents->count = 0;
 	contents->array =
-		(ArchiveContentItem *) calloc(lineCount, sizeof(ArchiveContentItem));
+		(ArchiveContentItem *) calloc(lbuf.count, sizeof(ArchiveContentItem));
 
-	for (int lineNumber = 0; lineNumber < lineCount; lineNumber++)
+	for (uint64_t lineNumber = 0; lineNumber < lbuf.count; lineNumber++)
 	{
 		ArchiveContentItem *item = &(contents->array[contents->count]);
 
-		char *line = lines[lineNumber];
+		char *line = lbuf.lines[lineNumber];
 
 		/* skip empty lines and lines that start with a semi-colon (comment) */
 		if (line == NULL || *line == '\0' || *line == ';')
@@ -1195,9 +1132,9 @@ parse_archive_list(const char *filename, ArchiveContentArray *contents)
 
 		if (!parse_archive_list_entry(item, line))
 		{
-			log_error("Failed to parse line %d of \"%s\", "
+			log_error("Failed to parse line %lld of \"%s\", "
 					  "see above for details",
-					  lineNumber,
+					  (long long) lineNumber,
 					  filename);
 			return false;
 		}
@@ -1310,6 +1247,14 @@ parse_archive_list_entry(ArchiveContentItem *item, const char *line)
 	}
 
 	item->desc = token.desc;
+	int itemDescLen = token.ptr - start + 1;
+
+	if (itemDescLen == 0)
+	{
+		log_error("Failed to parse Archive TOC: %s", line);
+		return false;
+	}
+
 	item->description = (char *) calloc(token.ptr - start + 1, sizeof(char));
 
 	if (item->description == NULL)
@@ -1359,6 +1304,13 @@ parse_archive_list_entry(ArchiveContentItem *item, const char *line)
 		/* 10. restore list name */
 		size_t len = strlen(token.ptr) + 1;
 		item->restoreListName = (char *) calloc(len, sizeof(char));
+
+		if (item->restoreListName == NULL)
+		{
+			log_error(ALLOCATION_FAILED_ERROR);
+			return false;
+		}
+
 		strlcpy(item->restoreListName, token.ptr, len);
 	}
 
@@ -1434,6 +1386,13 @@ tokenize_archive_list_entry(ArchiveToken *token)
 		int len = ptr - line + 1;
 		size_t size = len + 1;
 		char *buf = (char *) calloc(size, sizeof(char));
+
+		if (buf == NULL)
+		{
+			log_error(ALLOCATION_FAILED_ERROR);
+			return false;
+		}
+
 		strlcpy(buf, line, len);
 
 		if (!stringToUInt32(buf, &(token->oid)))
@@ -1594,27 +1553,3 @@ parse_archive_acl_or_comment(char *ptr, ArchiveContentItem *item)
 
 	return true;
 }
-
-
-/*
- * FreeArchiveContentArray frees the memory allocated in the given
- * ArchiveContentArray.
- */
-bool
-FreeArchiveContentArray(ArchiveContentArray *contents)
-{
-	for (int i = 0; i < contents->count; i++)
-	{
-		ArchiveContentItem *item = &(contents->array[i]);
-
-		free(item->description);
-		free(item->restoreListName);
-	}
-
-	free(contents->array);
-
-	contents->count = 0;
-	contents->array = NULL;
-
-	return true;
-}
