@@ -48,6 +48,7 @@ typedef enum
 	ARCHIVE_TAG_COMMENT,
 	ARCHIVE_TAG_CONSTRAINT,
 	ARCHIVE_TAG_CONVERSION,
+	ARCHIVE_TAG_DATABASE,
 	ARCHIVE_TAG_DEFAULT_ACL,
 	ARCHIVE_TAG_DEFAULT,
 	ARCHIVE_TAG_DOMAIN,
@@ -187,6 +188,7 @@ typedef struct RestoreOptions
 	bool noOwner;
 	bool noComments;
 	bool noACL;
+	bool noTableSpaces;
 	int jobs;
 } RestoreOptions;
 
@@ -239,6 +241,4 @@ bool parse_archive_acl_or_comment(char *ptr, ArchiveContentItem *item);
 bool parse_archive_list_entry(ArchiveContentItem *item, const char *line);
 bool tokenize_archive_list_entry(ArchiveToken *token);
 
-bool FreeArchiveContentArray(ArchiveContentArray *contents);
-
 #endif /* PGCMD_H */
