@@ -27,7 +27,6 @@
 
 
 static bool copydb_copy_supervisor_add_table_hook(void *ctx, SourceTable *table);
-static void FreeCopyArgs(CopyArgs *args);
 
 /*
  * copydb_table_data fetches the list of tables from the source database and
@@ -43,12 +42,6 @@ copydb_copy_all_table_data(CopyDataSpec *specs)
 {
 	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
 
-	if (!summary_start_timing(sourceDB, TIMING_SECTION_TOTAL_DATA))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
 	if (specs->runState.tableCopyIsDone &&
 		specs->runState.indexCopyIsDone &&
 		specs->runState.sequenceCopyIsDone &&
@@ -59,6 +52,12 @@ copydb_copy_all_table_data(CopyDataSpec *specs)
 		return true;
 	}
 
+	if (!summary_start_timing(sourceDB, TIMING_SECTION_TOTAL_DATA))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
 	/* close SQLite databases before fork() */
 	if (!catalog_close_from_specs(specs))
 	{
@@ -113,7 +112,9 @@ copydb_process_table_data(CopyDataSpec *specs)
 	/*
 	 * Take care of extensions configuration table in an auxilliary process.
 	 */
-	if (!copydb_start_extension_data_process(specs))
+	bool createExtension = false;
+
+	if (!copydb_start_extension_data_process(specs, createExtension))
 	{
 		/* errors have already been logged */
 		return false;
@@ -167,7 +168,7 @@ copydb_process_table_data(CopyDataSpec *specs)
 	}
 
 	/*
-	 * Are blobs table data? well pg_dump --section sayth yes.
+	 * Are blobs table data? well pg_dump --section says yes.
 	 */
 	if (errors == 0 && !copydb_start_blob_process(specs))
 	{
@@ -573,12 +574,23 @@ copydb_copy_supervisor_add_table_hook(void *ctx, SourceTable *table)
 		 * Before adding the table to be processed by workers, truncate it on
 		 * the target database now, avoiding concurrency issues.
 		 */
-		if (!pgsql_truncate(dst, table->qname))
+		bool granted = false;
+
+		if (!pgsql_has_table_privilege(dst, table->qname, "TRUNCATE", &granted))
 		{
 			/* errors have already been logged */
 			return false;
 		}
 
+		if (granted)
+		{
+			if (!pgsql_truncate(dst, table->qname))
+			{
+				/* errors have already been logged */
+				return false;
+			}
+		}
+
 		for (int i = 0; i < table->partition.partCount; i++)
 		{
 			int partNumber = i + 1;
@@ -738,7 +750,9 @@ copydb_table_data_worker(CopyDataSpec *specs)
 		return false;
 	}
 
-	while (true)
+	bool stop = false;
+
+	while (!stop)
 	{
 		QMessage mesg = { 0 };
 		bool recv_ok = queue_receive(&(specs->copyQueue), &mesg);
@@ -760,10 +774,9 @@ copydb_table_data_worker(CopyDataSpec *specs)
 		{
 			case QMSG_TYPE_STOP:
 			{
+				stop = true;
 				log_debug("Stop message received by COPY worker");
-				(void) copydb_close_snapshot(specs);
-				pgsql_finish(&dst);
-				return true;
+				break;
 			}
 
 			case QMSG_TYPE_TABLEPOID:
@@ -825,7 +838,7 @@ copydb_table_data_worker(CopyDataSpec *specs)
 		return false;
 	}
 
-	return errors == 0;
+	return stop == true && errors == 0;
 }
 
 
@@ -853,7 +866,6 @@ copydb_copy_data_by_oid(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 				  "see above for details",
 				  oid);
 
-		free(table);
 		return false;
 	}
 
@@ -869,7 +881,6 @@ copydb_copy_data_by_oid(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 	if (!copydb_init_table_specs(tableSpecs, specs, table, part))
 	{
 		/* errors have already been logged */
-		FreeCopyTableDataSpec(tableSpecs);
 		return false;
 	}
 
@@ -878,6 +889,25 @@ copydb_copy_data_by_oid(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 			  table->qname,
 			  part);
 
+	/*
+	 * Now check that the table still exists on the source server.
+	 */
+	bool tableStillExists = true;
+
+	if (!copydb_check_table_exists(src, table, &tableStillExists))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	if (!tableStillExists)
+	{
+		log_warn("Skipping table %s (oid %u) which does not exists anymore "
+				 "on the source database",
+				 table->qname, oid);
+		return true;
+	}
+
 	char psTitle[BUFSIZE] = { 0 };
 
 	if (table->partition.partCount > 0)
@@ -899,42 +929,44 @@ copydb_copy_data_by_oid(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 	 */
 	bool isDone = false;
 
-	if (!copydb_table_create_lockfile(specs, tableSpecs, &isDone))
+	if (!copydb_table_create_lockfile(specs, tableSpecs, dst, &isDone))
 	{
 		/* errors have already been logged */
-		FreeCopyTableDataSpec(tableSpecs);
 		return false;
 	}
 
-	if (isDone)
-	{
-		log_info("Skipping table %s (%u), already done on a previous run",
-				 tableSpecs->sourceTable->qname,
-				 tableSpecs->sourceTable->oid);
-		FreeCopyTableDataSpec(tableSpecs);
-		return true;
-	}
-
 	/*
-	 * 1. Now COPY the TABLE DATA from the source to the destination.
+	 * Skip only table-data copy when it it has been done already on a previous
+	 * run. We still need to process the indexes, constraints, and vacuum.
+	 * So, signal the index and vacuum workers as usual.
 	 */
-	if (!table->excludeData)
+	if (isDone)
 	{
-		if (!copydb_copy_table(specs, src, dst, tableSpecs))
+		log_info("Skipping table-data %s (%u), already done on a previous run",
+				 tableSpecs->sourceTable->qname,
+				 tableSpecs->sourceTable->oid);
+	}
+	else
+	{
+		/*
+		 * 1. Now COPY the TABLE DATA from the source to the destination.
+		 */
+		if (!table->excludeData)
+		{
+			if (!copydb_copy_table(specs, src, dst, tableSpecs))
+			{
+				/* errors have already been logged */
+				return false;
+			}
+		}
+
+		if (!copydb_mark_table_as_done(specs, tableSpecs))
 		{
 			/* errors have already been logged */
-			FreeCopyTableDataSpec(tableSpecs);
 			return false;
 		}
 	}
 
-	if (!copydb_mark_table_as_done(specs, tableSpecs))
-	{
-		/* errors have already been logged */
-		FreeCopyTableDataSpec(tableSpecs);
-		return false;
-	}
-
 	if (specs->section == DATA_SECTION_TABLE_DATA)
 	{
 		log_debug("Skip indexes, constraints, vacuum (section: table-data)");
@@ -965,7 +997,6 @@ copydb_copy_data_by_oid(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 											 &indexesAreBeingProcessed))
 		{
 			/* errors have already been logged */
-			FreeCopyTableDataSpec(tableSpecs);
 			return false;
 		}
 		else if (allPartsDone && !indexesAreBeingProcessed)
@@ -985,7 +1016,6 @@ copydb_copy_data_by_oid(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 			{
 				log_error("Failed to count indexes attached to table %s",
 						  tableSpecs->sourceTable->qname);
-				FreeCopyTableDataSpec(tableSpecs);
 				return false;
 			}
 
@@ -1000,7 +1030,6 @@ copydb_copy_data_by_oid(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 						log_error("Failed to queue VACUUM ANALYZE %s [%u]",
 								  sourceTable->qname,
 								  sourceTable->oid);
-						FreeCopyTableDataSpec(tableSpecs);
 						return false;
 					}
 				}
@@ -1010,14 +1039,11 @@ copydb_copy_data_by_oid(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 				log_error("Failed to add the indexes for %s, "
 						  "see above for details",
 						  tableSpecs->sourceTable->qname);
-				FreeCopyTableDataSpec(tableSpecs);
 				return false;
 			}
 		}
 	}
 
-	FreeCopyTableDataSpec(tableSpecs);
-
 	return true;
 }
 
@@ -1030,6 +1056,7 @@ copydb_copy_data_by_oid(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 bool
 copydb_table_create_lockfile(CopyDataSpec *specs,
 							 CopyTableDataSpec *tableSpecs,
+							 PGSQL *dst,
 							 bool *isDone)
 {
 	DatabaseCatalog *sourceDB = &(specs->catalogs.source);
@@ -1088,8 +1115,8 @@ copydb_table_create_lockfile(CopyDataSpec *specs,
 		}
 	}
 
-	/* COPY FROM tablename, or maybe COPY FROM (SELECT ... WHERE ...) */
-	if (!catalog_s_table_fetch_attrs(sourceDB, tableSpecs->sourceTable))
+	/* build the table attributes' list */
+	if (!catalog_s_table_attrlist(sourceDB, tableSpecs->sourceTable))
 	{
 		log_error("Failed to fetch table %s attribute list, "
 				  "see above for details",
@@ -1097,17 +1124,45 @@ copydb_table_create_lockfile(CopyDataSpec *specs,
 		return false;
 	}
 
+	/* COPY FROM tablename, or maybe COPY FROM (SELECT ... WHERE ...) */
 	CopyArgs *args = &(tableSpecs->copyArgs);
 
 	args->srcQname = tableSpecs->sourceTable->qname;
-	args->srcAttrList = NULL;
+	args->srcAttrList = tableSpecs->sourceTable->attrList;
 	args->srcWhereClause = NULL;
 	args->dstQname = tableSpecs->sourceTable->qname;
-	args->dstAttrList = NULL;
-	args->truncate = tableSpecs->sourceTable->partition.partCount <= 1;
+	args->dstAttrList = tableSpecs->sourceTable->attrList;
+	args->truncate = false;     /* default value, see below */
 	args->freeze = tableSpecs->sourceTable->partition.partCount <= 1;
 	args->bytesTransmitted = 0;
 
+	/*
+	 * Check to see if we want to TRUNCATE the table and benefit from the COPY
+	 * FREEZE optimisation.
+	 *
+	 * First, if the table COPY is partitionned then we truncate at the
+	 * top-level rather than for each partition, disabling the COPY FREEZE
+	 * optimisation.
+	 *
+	 * Second, we need the permission to run the TRUNCATE command on the target
+	 * table on the target database.
+	 */
+	if (tableSpecs->sourceTable->partition.partCount <= 1)
+	{
+		bool granted = false;
+
+		if (!pgsql_has_table_privilege(dst,
+									   tableSpecs->sourceTable->qname,
+									   "TRUNCATE",
+									   &granted))
+		{
+			/* errors have already been logged */
+			return false;
+		}
+
+		args->truncate = granted;
+	}
+
 	if (!copydb_prepare_copy_query(tableSpecs, args))
 	{
 		/* errors have already been logged */
@@ -1256,7 +1311,6 @@ copydb_copy_table(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 
 	/* Now copy the data from source to target */
 	CopyTableSummary *summary = &(tableSpecs->summary);
-	log_notice("%s", summary->command);
 
 	int attempts = 0;
 	int maxAttempts = 5;        /* allow 5 attempts total, 4 retries */
@@ -1333,17 +1387,6 @@ copydb_copy_table(CopyDataSpec *specs, PGSQL *src, PGSQL *dst,
 bool
 copydb_prepare_copy_query(CopyTableDataSpec *tableSpecs, CopyArgs *args)
 {
-	PQExpBuffer srcAttrList = createPQExpBuffer();
-
-	if (!copydb_prepare_copy_query_attrlist(tableSpecs, srcAttrList))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	args->srcAttrList = strdup(srcAttrList->data);
-	destroyPQExpBuffer(srcAttrList);
-
 	/*
 	 * On a source COPY query we might want to add filtering.
 	 */
@@ -1363,7 +1406,7 @@ copydb_prepare_copy_query(CopyTableDataSpec *tableSpecs, CopyArgs *args)
 				/* the last part for ctid splits covers "extra" relpages */
 				appendPQExpBuffer(srcWhereClause,
 								  "WHERE ctid >= '(%lld,0)'::tid",
-								  (long long) tableSpecs->part.min + 1);
+								  (long long) tableSpecs->part.min);
 			}
 			else
 			{
@@ -1376,11 +1419,31 @@ copydb_prepare_copy_query(CopyTableDataSpec *tableSpecs, CopyArgs *args)
 		}
 		else
 		{
-			appendPQExpBuffer(srcWhereClause,
-							  "WHERE %s BETWEEN %lld AND %lld",
-							  tableSpecs->part.partKey,
-							  (long long) tableSpecs->part.min,
-							  (long long) tableSpecs->part.max);
+			/* partition to take care of NULL values */
+			if (tableSpecs->part.min == -1 &&
+				tableSpecs->part.max == -1)
+			{
+				appendPQExpBuffer(srcWhereClause,
+								  "WHERE %s IS NULL",
+								  tableSpecs->part.partKey);
+			}
+
+			/* the last partition has no upper bound */
+			else if (tableSpecs->part.max == -1)
+			{
+				appendPQExpBuffer(srcWhereClause,
+								  "WHERE %s >= %lld",
+								  tableSpecs->part.partKey,
+								  (long long) tableSpecs->part.min);
+			}
+			else
+			{
+				appendPQExpBuffer(srcWhereClause,
+								  "WHERE %s BETWEEN %lld AND %lld",
+								  tableSpecs->part.partKey,
+								  (long long) tableSpecs->part.min,
+								  (long long) tableSpecs->part.max);
+			}
 		}
 
 		if (PQExpBufferBroken(srcWhereClause))
@@ -1394,69 +1457,6 @@ copydb_prepare_copy_query(CopyTableDataSpec *tableSpecs, CopyArgs *args)
 		destroyPQExpBuffer(srcWhereClause);
 	}
 
-	/*
-	 * Prepare COPY args for destination query (COPY ... FROM)
-	 */
-	PQExpBuffer dstAttrList = createPQExpBuffer();
-
-	if (!copydb_prepare_copy_query_attrlist(tableSpecs, dstAttrList))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	args->dstAttrList = strdup(dstAttrList->data);
-	destroyPQExpBuffer(dstAttrList);
-
-	return true;
-}
-
-
-/*
- * copydb_prepare_copy_query_attrlist prepares the attribute list from a given
- * table specification.
- */
-bool
-copydb_prepare_copy_query_attrlist(CopyTableDataSpec *tableSpecs,
-								   PQExpBuffer attrList)
-{
-	SourceTable *table = tableSpecs->sourceTable;
-
-	bool isFirst = true;
-
-	for (int i = 0; i < table->attributes.count; i++)
-	{
-		SourceTableAttribute *attribute = &(table->attributes.array[i]);
-		char *attname = attribute->attname;
-
-		/* Generated columns cannot be used in COPY */
-		if (attribute->attisgenerated)
-		{
-			log_debug("Skipping %s.%s in COPY as it is a generated column",
-					  tableSpecs->sourceTable->qname,
-					  attname);
-			continue;
-		}
-
-		if (isFirst)
-		{
-			isFirst = false;
-		}
-		else
-		{
-			appendPQExpBufferStr(attrList, ", ");
-		}
-
-		appendPQExpBuffer(attrList, "%s", attname);
-	}
-
-	if (PQExpBufferBroken(attrList))
-	{
-		log_error("Failed to create attribute list for %s: out of memory",
-				  tableSpecs->sourceTable->qname);
-		return false;
-	}
-
 	return true;
 }
 
@@ -1483,6 +1483,9 @@ copydb_prepare_summary_command(CopyTableDataSpec *tableSpecs)
 	tableSummary->command =
 		command->data != NULL ? strdup(command->data) : NULL;
 
+	/* also keep a pointer around in the copyArgs structure */
+	tableSpecs->copyArgs.logCommand = tableSummary->command;
+
 	if (PQExpBufferBroken(command) || tableSummary->command == NULL)
 	{
 		log_error("Failed to create summary command for %s: out of memory",
@@ -1495,25 +1498,50 @@ copydb_prepare_summary_command(CopyTableDataSpec *tableSpecs)
 
 
 /*
- * FreeCopyTableDataSpec takes care of free'ing the allocated memory for the
- * CopyTableDataSpec.
+ * copydb_check_table_exists checks that a table still exists. In order to
+ * avoid race conditions when checking for existence, grab a explicit ACCESS
+ * SHARE LOCK on the table.
  */
-void
-FreeCopyTableDataSpec(CopyTableDataSpec *tableSpecs)
+bool
+copydb_check_table_exists(PGSQL *pgsql, SourceTable *table, bool *exists)
 {
-	free(tableSpecs->sourceTable);
-	free(tableSpecs->summary.command);
-	FreeCopyArgs(&(tableSpecs->copyArgs));
-}
+	if (!pgsql_table_exists(pgsql,
+							table->oid,
+							table->nspname,
+							table->relname,
+							exists))
+	{
+		/* errors have already been logged */
+		return false;
+	}
 
+	/* if the table does not exists, we stop here */
+	if (!exists)
+	{
+		return true;
+	}
 
-/*
- * FreeCopyArgs takes care of free'ing the allocated memory for the CopyArgs.
- */
-static void
-FreeCopyArgs(CopyArgs *args)
-{
-	free(args->srcAttrList);
-	free(args->srcWhereClause);
-	free(args->dstAttrList);
+	/* if the table was reported to exists, try and lock it */
+	bool locked = pgsql_lock_table(pgsql, table->qname, "ACCESS SHARE");
+
+	if (!locked)
+	{
+		log_error("Failed to LOCK table %s in ACCESS SHARE mode", table->qname);
+	}
+
+	/*
+	 * If we failed to obtain the lock, maybe the table doesn't exists anymore,
+	 * in which case we do not want to report an error condition.
+	 */
+	if (!pgsql_table_exists(pgsql,
+							table->oid,
+							table->nspname,
+							table->relname,
+							exists))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	return locked || !(*exists);
 }
