@@ -30,9 +30,26 @@ EOF
 PAGILA_SOURCE_PGURI="postgres://pagila:0wn3d@source/pagila"
 PAGILA_TARGET_PGURI="postgres://pagila:0wn3d@target/pagila"
 
+# Create extensions in the default and alternative schemas to always
+# ensure these copy as expected.
 psql -d ${PAGILA_SOURCE_PGURI} <<EOF
 create extension ltree;
 create extension hstore;
+create schema foo;
+create extension intarray schema foo cascade;
+create schema bar;
+create extension pg_trgm schema bar cascade;
+create schema "Foo Bar";
+create extension pgcrypto schema "Foo Bar" cascade;
+create schema "With a space";
+create extension seg schema "With a space" cascade;
+EOF
+
+# Create a couple of schemas in the target database
+# to ensure they get skipped during the clone
+psql -d ${PAGILA_TARGET_PGURI} <<EOF
+create schema bar;
+create schema "With a space";
 EOF
 
 grep -v "OWNER TO postgres" /usr/src/pagila/pagila-schema.sql > /tmp/pagila-schema.sql
@@ -40,7 +57,9 @@ grep -v "OWNER TO postgres" /usr/src/pagila/pagila-schema.sql > /tmp/pagila-sche
 psql -o /tmp/s.out -d ${PAGILA_SOURCE_PGURI} -1 -f /tmp/pagila-schema.sql
 psql -o /tmp/d.out -d ${PAGILA_SOURCE_PGURI} -1 -f /usr/src/pagila/pagila-data.sql
 
-pgcopydb clone --skip-ext-comments --notice \
+pgcopydb clone --notice \
+         --skip-ext-comments \
+         --skip-db-properties \
          --source ${PAGILA_SOURCE_PGURI} \
          --target ${PAGILA_TARGET_PGURI}
 
